"""
Optimal scaling
===============

Contains the implementation of the optimal scaling approach, applied for integration
of ordinal and censored data in ODE modeling. Each ordinal datapoint is defined as
being part of a category, where the mutual ordering of the categories of the same observable
is known. The category interval bounds are numerically optimized and quantitative surrogate
measurements are calculated to represent the ordinal measurements. This constitutes
the inner subproblem of the hierarchical optimization problem. For censored data, as the
category interval bounds are known, the surrogate measurements are directly calculated.

An example of parameter estimation with ordinal data can be found in pypesto/doc/examples/example_ordinal.ipynb.
An example of parameter estimation with censored data can be found in pypesto/doc/examples/example_censored.ipynb.

Details on the optimal scaling approach can be found in Shepard, 1962 (https://doi.org/10.1007/BF02289621).
Details on the application of the gradient-based optimal scaling approach to mechanistic modeling
with ordinal data can be found in Schmiester et al. 2020 (https://doi.org/10.1007/s00285-020-01522-w)
and Schmiester et al. 2021 (https://doi.org/10.1093/bioinformatics/btab512).
"""

from .calculator import OptimalScalingAmiciCalculator
from .parameter import OptimalScalingParameter
from .problem import OptimalScalingProblem
from .solver import OptimalScalingInnerSolver



from typing import Dict, List, Tuple

import numpy as np
import pandas as pd

from ...C import (
    CURRENT_SIMULATION,
    DATAPOINTS,
    EXPDATA_MASK,
    INNER_PARAMETER_BOUNDS,
    LIN,
    MAX_DATAPOINT,
    MEASUREMENT_TYPE,
    MIN_DATAPOINT,
    N_SPLINE_PARS,
    NOISE_PARAMETERS,
    NONLINEAR_MONOTONE,
    NUM_DATAPOINTS,
    SPLINE_PAR_TYPE,
    TIME,
    InnerParameterType,
)
from ..problem import (
    InnerProblem,
    _get_timepoints_with_replicates,
    ix_matrices_from_arrays,
)
from .parameter import SplineInnerParameter

try:
    import amici
    import petab
    from petab.C import OBSERVABLE_ID
except ImportError:
    pass


class SplineInnerProblem(InnerProblem):
    """Inner optimization problem for spline approximation.

    Attributes
    ----------
    xs:
        Mapping of (inner) parameter ID to ``InnerParameters``.
    data:
        Measurement data. One matrix (`num_timepoints` x `num_observables`)
        per simulation condition. Missing observations as NaN.
    groups:
        A dictionary of the groups of the subproblem.
    spline_ratio:
        The ratio of the number of spline inner parameters and number of measurements for each group.
    """

    def __init__(
        self,
        xs: List[SplineInnerParameter],
        data: List[np.ndarray],
        spline_ratio: float = 0.5,
    ):
        """Construct."""
        super().__init__(xs, data)
        self.spline_ratio = spline_ratio

        if spline_ratio <= 0:
            raise ValueError("Spline ratio must be a positive float.")
        self._initialize_groups()

    def _initialize_groups(self) -> None:
        """Initialize the groups of the subproblem."""
        self.groups = {}
        for group in self.get_groups_for_xs(InnerParameterType.SPLINE):
            xs = self.get_xs_for_group(group)

            self.groups[group] = {}
            self.groups[group][N_SPLINE_PARS] = len({x.index for x in xs})
            self.groups[group][DATAPOINTS] = self.get_measurements_for_group(
                group
            )
            self.groups[group][NUM_DATAPOINTS] = len(
                self.groups[group][DATAPOINTS]
            )
            self.groups[group][MIN_DATAPOINT] = np.min(
                self.groups[group][DATAPOINTS]
            )
            self.groups[group][MAX_DATAPOINT] = np.max(
                self.groups[group][DATAPOINTS]
            )

            self.groups[group][EXPDATA_MASK] = xs[0].ixs
            self.groups[group][CURRENT_SIMULATION] = np.zeros(
                self.groups[group][NUM_DATAPOINTS]
            )
            self.groups[group][NOISE_PARAMETERS] = np.zeros(
                self.groups[group][NUM_DATAPOINTS]
            )

    @staticmethod
    def from_petab_amici(
        petab_problem: petab.Problem,
        amici_model: 'amici.Model',
        edatas: List['amici.ExpData'],
        spline_ratio: float = None,
    ) -> 'SplineInnerProblem':
        """Construct the inner problem from the `petab_problem`."""
        if spline_ratio is None:
            spline_ratio = get_default_options()
        return spline_inner_problem_from_petab_problem(
            petab_problem, amici_model, edatas, spline_ratio
        )

    def get_groups_for_xs(self, inner_parameter_type: str) -> List[int]:
        """Get unique list of ``SplineParameter.group`` values."""
        groups = [x.group for x in self.get_xs_for_type(inner_parameter_type)]
        return list(set(groups))

    def get_xs_for_group(self, group: int) -> List[SplineInnerParameter]:
        r"""Get ``SplineParameter``\s that belong to the given group."""
        return [x for x in self.xs.values() if x.group == group]

    def get_free_xs_for_group(self, group: int) -> List[SplineInnerParameter]:
        r"""Get ``SplineParameter``\s that are free and belong to the given group."""
        return [
            x
            for x in self.xs.values()
            if x.group == group and x.estimate is True
        ]

    def get_fixed_xs_for_group(self, group: int) -> List[SplineInnerParameter]:
        r"""Get ``SplineParameter``\s that are fixed and belong to the given group."""
        return [
            x
            for x in self.xs.values()
            if x.group == group and x.estimate is False
        ]

    def get_inner_parameter_dictionary(self) -> Dict:
        """Get a dictionary with all inner parameter ids and their values."""
        inner_par_dict = {}
        for x_id, x in self.xs.items():
            inner_par_dict[x_id] = x.value
        return inner_par_dict

    def get_measurements_for_group(self, gr) -> np.ndarray:
        """Get measurements for a group."""
        # Taking the ixs of first inner parameter since
        # all of them are the same for the same group.
        ixs = self.get_xs_for_group(gr)[0].ixs

        return np.concatenate(
            [
                self.data[condition_index][ixs[condition_index]]
                for condition_index in range(len(ixs))
            ]
        )


def get_default_options() -> Dict:
    """Return the default spline problem options dictionary."""
    spline_ratio = 1 / 2
    return spline_ratio


def spline_inner_problem_from_petab_problem(
    petab_problem: petab.Problem,
    amici_model: 'amici.Model',
    edatas: List['amici.ExpData'],
    spline_ratio: float = None,
):
    """Construct the inner problem from the `petab_problem`."""
    if spline_ratio is None:
        spline_ratio = get_default_options()
    elif spline_ratio <= 0:
        raise ValueError("Spline ratio must be a positive float.")

    # inner parameters
    inner_parameters = spline_inner_parameters_from_measurement_df(
        petab_problem.measurement_df, spline_ratio, amici_model
    )

    # used indices for all measurement specific parameters
    ixs = spline_ixs_for_measurement_specific_parameters(
        petab_problem, amici_model, inner_parameters
    )

    # transform experimental data
    edatas = [
        amici.numpy.ExpDataView(edata)['observedData'] for edata in edatas
    ]

    # matrixify
    ix_matrices = ix_matrices_from_arrays(ixs, edatas)

    # assign matrices
    for par in inner_parameters:
        par.ixs = ix_matrices[par.inner_parameter_id]

    return SplineInnerProblem(
        inner_parameters,
        edatas,
        spline_ratio,
    )


def spline_inner_parameters_from_measurement_df(
    df: pd.DataFrame,
    spline_ratio: float,
    amici_model: 'amici.Model',
) -> List[SplineInnerParameter]:
    """Create list of inner free parameters from PEtab measurement table."""
    df = df.reset_index()

    observable_ids = amici_model.getObservableIds()

    par_type = SPLINE_PAR_TYPE
    estimate = True
    lb, ub = INNER_PARAMETER_BOUNDS[InnerParameterType.SPLINE].values()

    inner_parameters = []

    # Select the nonlinear monotone measurements.
    df = df[df[MEASUREMENT_TYPE] == NONLINEAR_MONOTONE]

    # Iterate over groups.
    for observable_id in observable_ids:
        group = observable_ids.index(observable_id) + 1
        df_for_group = df[df[OBSERVABLE_ID] == observable_id]

        n_spline_parameters = int(np.ceil(len(df_for_group) * spline_ratio))

        # Create n_spline_parameters number of spline inner parameters.
        for par_index in range(n_spline_parameters):
            par_id = f'{par_type}_{observable_id}_{group}_{par_index+1}'
            inner_parameters.append(
                SplineInnerParameter(
                    inner_parameter_id=par_id,
                    inner_parameter_type=InnerParameterType.SPLINE,
                    scale=LIN,
                    lb=lb,
                    ub=ub,
                    observable_id=observable_id,
                    group=group,
                    index=par_index + 1,
                    estimate=estimate,
                )
            )

    inner_parameters.sort(key=lambda x: (x.group, x.index))

    return inner_parameters


def spline_ixs_for_measurement_specific_parameters(
    petab_problem: 'petab.Problem',
    amici_model: 'amici.Model',
    inner_parameters: List[SplineInnerParameter],
) -> Dict[str, List[Tuple[int, int, int]]]:
    """Create mapping of parameters to measurements.

    Returns
    -------
    A dictionary mapping parameter ID to a list of
    `(condition index, time index, observable index)` tuples in which this
    output parameter is used. For each condition, the time index refers to
    a sorted list of non-unique time points for which there are measurements.
    """
    ixs_for_par = {}
    observable_ids = amici_model.getObservableIds()

    simulation_conditions = (
        petab_problem.get_simulation_conditions_from_measurement_df()
    )
    for condition_ix, condition in simulation_conditions.iterrows():
        # measurement table for current condition
        df_for_condition = petab.get_rows_for_condition(
            measurement_df=petab_problem.measurement_df, condition=condition
        )

        # unique sorted list of timepoints
        timepoints = sorted(df_for_condition[TIME].unique().astype(float))
        # non-unique sorted list of timepoints
        timepoints_w_reps = _get_timepoints_with_replicates(
            measurement_df=df_for_condition
        )

        for time in timepoints:
            # subselect measurements for time `time`
            df_for_time = df_for_condition[df_for_condition[TIME] == time]
            time_ix_0 = timepoints_w_reps.index(time)

            # remember used time indices for each observable
            time_ix_for_obs_ix = {}

            # iterate over measurements
            for _, measurement in df_for_time.iterrows():
                # extract observable index
                observable_ix = observable_ids.index(
                    measurement[OBSERVABLE_ID]
                )

                # as the time indices have to account for replicates, we need
                #  to track which time indices have already been assigned for
                #  the current observable
                if observable_ix in time_ix_for_obs_ix:
                    # a replicate
                    time_ix_for_obs_ix[observable_ix] += 1
                else:
                    # the first measurement for this `(observable, timepoint)`
                    time_ix_for_obs_ix[observable_ix] = time_ix_0
                time_w_reps_ix = time_ix_for_obs_ix[observable_ix]

                inner_par_ids_for_meas = (
                    get_spline_inner_par_ids_for_measurement(
                        measurement, inner_parameters
                    )
                )

                # try to insert if hierarchical parameter
                for override in inner_par_ids_for_meas:
                    ixs_for_par.setdefault(override, []).append(
                        (condition_ix, time_w_reps_ix, observable_ix)
                    )
    return ixs_for_par


def get_spline_inner_par_ids_for_measurement(
    measurement: Dict,
    inner_parameters: List[SplineInnerParameter],
):
    """Return inner parameter ids of parameters which are related to the measurement."""
    return [
        inner_par.inner_parameter_id
        for inner_par in inner_parameters
        if inner_par.observable_id == measurement[OBSERVABLE_ID]
    ]



"""Engines with multi-process parallelization."""
import logging
import multiprocessing
import os
from typing import Any, List

import cloudpickle as pickle
from tqdm import tqdm

from .base import Engine
from .task import Task

logger = logging.getLogger(__name__)


def work(pickled_task):
    """Unpickle and execute task."""
    task = pickle.loads(pickled_task)
    return task.execute()


class MultiProcessEngine(Engine):
    """
    Parallelize the task execution using multiprocessing.

    Parameters
    ----------
    n_procs:
        The maximum number of processes to use in parallel.
        Defaults to the number of CPUs available on the system according to
        `os.cpu_count()`.
        The effectively used number of processes will be the minimum of
        `n_procs` and the number of tasks submitted.
    method:
        Start method, any of "fork", "spawn", "forkserver", or None,
        giving the system specific default context.
    """

    def __init__(self, n_procs: int = None, method: str = None):
        super().__init__()

        if n_procs is None:
            n_procs = os.cpu_count()
            logger.info(
                f"Engine will use up to {n_procs} processes (= CPU count)."
            )
        self.n_procs: int = n_procs
        self.method: str = method

    def execute(
        self, tasks: List[Task], progress_bar: bool = True
    ) -> List[Any]:
        """Pickle tasks and distribute work over parallel processes.

        Parameters
        ----------
        tasks:
            List of tasks to execute.
        progress_bar:
            Whether to display a progress bar.
        """
        n_tasks = len(tasks)

        pickled_tasks = [pickle.dumps(task) for task in tasks]

        n_procs = min(self.n_procs, n_tasks)
        logger.debug(f"Parallelizing on {n_procs} processes.")

        ctx = multiprocessing.get_context(method=self.method)

        with ctx.Pool(processes=n_procs) as pool:
            results = list(
                tqdm(
                    pool.imap(work, pickled_tasks),
                    total=len(pickled_tasks),
                    disable=not progress_bar,
                ),
            )

        return results



import copy
from typing import Dict, List, Sequence, Tuple

import numpy as np

from ...C import (
    AMICI_SIGMAY,
    AMICI_SY,
    AMICI_Y,
    FVAL,
    GRAD,
    HESS,
    MODE_RES,
    RDATAS,
    RES,
    SRES,
    X_INNER_OPT,
)
from ...objective.amici.amici_calculator import (
    AmiciCalculator,
    AmiciModel,
    AmiciSolver,
)
from ...objective.amici.amici_util import (
    filter_return_dict,
    init_return_values,
)
from .problem import SplineInnerProblem
from .solver import SplineInnerSolver

try:
    import amici
    from amici.parameter_mapping import ParameterMapping
except ImportError:
    pass


class SplineAmiciCalculator(AmiciCalculator):
    """A calculator is passed as `calculator` to the pypesto.AmiciObjective.

    The object is called by :func:`pypesto.AmiciObjective.call_unprocessed`
    to calculate the current objective function values and gradient.
    """

    def __init__(
        self,
        inner_problem: SplineInnerProblem,
        inner_solver: SplineInnerSolver = None,
    ):
        """Initialize the calculator from the given problem.

        Parameters
        ----------
        inner_problem:
            The optimal scaling inner problem.
        inner_solver:
            A solver to solve ``inner_problem``.
            Defaults to ``SplineInnerSolver``.
        """
        super().__init__()
        self.inner_problem = inner_problem

        if inner_solver is None:
            inner_solver = SplineInnerSolver()
        self.inner_solver = inner_solver

    def initialize(self):
        """Initialize."""
        self.inner_solver.initialize()

    def __call__(
        self,
        x_dct: Dict,
        sensi_orders: Tuple[int, ...],
        mode: str,
        amici_model: AmiciModel,
        amici_solver: AmiciSolver,
        edatas: List['amici.ExpData'],
        n_threads: int,
        x_ids: Sequence[str],
        parameter_mapping: ParameterMapping,
        fim_for_hess: bool,
        rdatas: List['amici.ReturnData'] = None,
    ):
        """Perform the actual AMICI call.

        Parameters
        ----------
        x_dct:
            Parameters for which to compute function value and derivatives.
        sensi_orders:
            Tuple of requested sensitivity orders.
        mode:
            Call mode (function value or residual based).
        amici_model:
            The AMICI model.
        amici_solver:
            The AMICI solver.
        edatas:
            The experimental data.
        n_threads:
            Number of threads for AMICI call.
        x_ids:
            Ids of optimization parameters.
        parameter_mapping:
            Mapping of optimization to simulation parameters.
        fim_for_hess:
            Whether to use the FIM (if available) instead of the Hessian (if
            requested).
        inner_rdatas:
            AMICI simulation return data. In case the calculator is part of
            the :class:`pypesto.objective.amici.InnerCalculatorCollector`,
            it will already simulate the model and pass the results here.


        Returns
        -------
        inner_result:
            A dict containing the calculation results: FVAL, GRAD, RDATAS and X_INNER_OPT.
        """
        if mode == MODE_RES:
            raise ValueError(
                "Spline approximation cannot be called with residual mode."
            )
        if 2 in sensi_orders:
            raise ValueError(
                "Hessian and FIM are not implemented for the spline calculator."
            )

        # get dimension of outer problem
        dim = len(x_ids)

        # initialize return values
        nllh, snllh, s2nllh, chi2, res, sres = init_return_values(
            sensi_orders, mode, dim
        )
        # set order in solver
        sensi_order = 0
        if sensi_orders:
            sensi_order = max(sensi_orders)

        # If AMICI ReturnData is not provided, we need to simulate the model
        if rdatas is None:
            amici_solver.setSensitivityOrder(sensi_order)

            x_dct = copy.deepcopy(x_dct)

            # fill in parameters
            amici.parameter_mapping.fill_in_parameters(
                edatas=edatas,
                problem_parameters=x_dct,
                scaled_parameters=True,
                parameter_mapping=parameter_mapping,
                amici_model=amici_model,
            )
            # run amici simulation
            rdatas = amici.runAmiciSimulations(
                amici_model,
                amici_solver,
                edatas,
                num_threads=min(n_threads, len(edatas)),
            )

        inner_result = {
            FVAL: nllh,
            GRAD: snllh,
            HESS: s2nllh,
            RES: res,
            SRES: sres,
            RDATAS: rdatas,
            X_INNER_OPT: self.inner_problem.get_inner_parameter_dictionary(),
        }

        # if any amici simulation failed, it's unlikely we can compute
        # meaningful inner parameters, so we better just fail early.
        if any(rdata.status != amici.AMICI_SUCCESS for rdata in rdatas):
            inner_result[FVAL] = np.inf
            # if the gradient was requested,
            # we need to provide some value for it
            if 1 in sensi_orders:
                inner_result[GRAD] = np.full(
                    shape=len(x_ids), fill_value=np.nan
                )
            return filter_return_dict(inner_result)

        sim = [rdata[AMICI_Y] for rdata in rdatas]
        sigma = [rdata[AMICI_SIGMAY] for rdata in rdatas]

        # Clip negative simulation values to zero, to avoid numerical issues.
        for i in range(len(sim)):
            sim[i] = sim[i].clip(min=0)

        # compute optimal inner parameters
        x_inner_opt = self.inner_solver.solve(self.inner_problem, sim, sigma)
        inner_result[FVAL] = self.inner_solver.calculate_obj_function(
            x_inner_opt
        )
        inner_result[
            X_INNER_OPT
        ] = self.inner_problem.get_inner_parameter_dictionary()

        # Calculate analytical gradients if requested
        if sensi_order > 0:
            sy = [rdata[AMICI_SY] for rdata in rdatas]
            inner_result[GRAD] = self.inner_solver.calculate_gradients(
                problem=self.inner_problem,
                x_inner_opt=x_inner_opt,
                sim=sim,
                sigma=sigma,
                sy=sy,
                parameter_mapping=parameter_mapping,
                par_opt_ids=x_ids,
                par_sim_ids=amici_model.getParameterIds(),
                snllh=snllh,
            )

        return filter_return_dict(inner_result)



"""
Hierarchical
============

Hierarchical optimization sub-package.
"""

from .calculator import HierarchicalAmiciCalculator
from .parameter import InnerParameter
from .problem import InnerProblem
from .solver import AnalyticalInnerSolver, InnerSolver, NumericalInnerSolver



"""
Spline approximation
====================

Contains the implementation of a spline approximation approach, applied for integration
of nonlinear-monotone data in ODE modeling, where the data is assumed to have
an unknown monotone relationship with the model output. This relationship is
approximated by a piecewise linear (spline) function, which is numerically
optimized to fit the data. This constitutes the inner subproblem of the
hierarchical optimization problem.
"""

from .calculator import SplineAmiciCalculator
from .parameter import SplineInnerParameter
from .problem import SplineInnerProblem
from .solver import SplineInnerSolver



"""Definition of an optimal scaling parameter class."""
from typing import Dict, List, Tuple

import numpy as np
import pandas as pd

from ...C import (
    C_MATRIX,
    CAT_LB,
    CAT_UB,
    CENSORED,
    CENSORING_BOUNDS,
    CENSORING_TYPES,
    INNER_PARAMETER_BOUNDS,
    INTERVAL_CENSORED,
    LB_INDICES,
    LEFT_CENSORED,
    LIN,
    MEASUREMENT_CATEGORY,
    MEASUREMENT_TYPE,
    NUM_CATEGORIES,
    NUM_CONSTR_FULL,
    NUM_DATAPOINTS,
    NUM_INNER_PARAMS,
    ORDINAL,
    QUANTITATIVE_DATA,
    QUANTITATIVE_IXS,
    REDUCED,
    RIGHT_CENSORED,
    STANDARD,
    SURROGATE_DATA,
    TIME,
    UB_INDICES,
    W_DOT_MATRIX,
    W_MATRIX,
    InnerParameterType,
)
from ..problem import (
    InnerProblem,
    _get_timepoints_with_replicates,
    ix_matrices_from_arrays,
)
from .parameter import OptimalScalingParameter

try:
    import amici
    import petab
    from petab.C import OBSERVABLE_ID, PARAMETER_SEPARATOR
except ImportError:
    pass


class OptimalScalingProblem(InnerProblem):
    """Inner optimization problem for optimal scaling.

    Attributes
    ----------
    xs:
        Mapping of (inner) parameter ID to ``InnerParameters``.
    data:
        Measurement data. One matrix (`num_timepoints` x `num_observables`)
        per simulation condition. Missing observations as NaN.
    groups:
        A dictionary of the groups of the subproblem.
    method:
        A string representing the method of the Optimal Scaling approach, either 'reduced' or 'standard'.
    """

    def __init__(
        self,
        xs: List[OptimalScalingParameter],
        data: List[np.ndarray],
        method: str,
    ):
        """Construct."""
        super().__init__(xs, data)
        self.groups = {}
        self.method = method

        self._initialize_groups()

    def _initialize_groups(self) -> None:
        """Initialize the groups of the subproblem."""
        self.groups = {}

        for group in self.get_groups_for_xs(
            InnerParameterType.OPTIMAL_SCALING
        ):
            self.groups[group] = {}
            xs = self.get_xs_for_group(group)
            self.groups[group][NUM_CATEGORIES] = len({x.category for x in xs})
            self.groups[group][NUM_DATAPOINTS] = sum(
                [
                    sum([np.sum(ixs) for ixs in x.ixs])
                    for x in self.get_cat_ub_parameters_for_group(group)
                ]
            )

            self.groups[group][SURROGATE_DATA] = np.zeros(
                self.groups[group][NUM_DATAPOINTS]
            )

            self.groups[group][NUM_INNER_PARAMS] = (
                self.groups[group][NUM_DATAPOINTS]
                + 2 * self.groups[group][NUM_CATEGORIES]
            )

            self.groups[group][LB_INDICES] = list(
                range(
                    self.groups[group][NUM_DATAPOINTS],
                    self.groups[group][NUM_DATAPOINTS]
                    + self.groups[group][NUM_CATEGORIES],
                )
            )

            self.groups[group][UB_INDICES] = list(
                range(
                    self.groups[group][NUM_DATAPOINTS]
                    + self.groups[group][NUM_CATEGORIES],
                    self.groups[group][NUM_INNER_PARAMS],
                )
            )

            if all(x.censoring_type is not None for x in xs):
                self.groups[group][MEASUREMENT_TYPE] = CENSORED
                self.groups[group][
                    QUANTITATIVE_IXS
                ] = self.get_censored_group_quantitative_ixs(xs)
                self.groups[group][QUANTITATIVE_DATA] = np.concatenate(
                    [
                        data_i[mask_i]
                        for data_i, mask_i in zip(
                            self.data, self.groups[group][QUANTITATIVE_IXS]
                        )
                    ]
                )
            elif all(x.censoring_type is None for x in xs):
                self.groups[group][MEASUREMENT_TYPE] = ORDINAL

                self.groups[group][NUM_CONSTR_FULL] = (
                    2 * self.groups[group][NUM_DATAPOINTS]
                    + 2 * self.groups[group][NUM_CATEGORIES]
                )

                self.groups[group][C_MATRIX] = self.initialize_c(group)

                self.groups[group][W_MATRIX] = self.initialize_w(group)

                self.groups[group][W_DOT_MATRIX] = self.initialize_w(group)
            else:
                raise ValueError(
                    'Censoring types of optimal scaling parameters of a group '
                    'have to either be all None, or all not None.'
                )

    @staticmethod
    def from_petab_amici(
        petab_problem: petab.Problem,
        amici_model: 'amici.Model',
        edatas: List['amici.ExpData'],
        method: str = None,
    ) -> 'OptimalScalingProblem':
        """Construct the inner problem from the `petab_problem`."""
        if not method:
            method = REDUCED

        return optimal_scaling_inner_problem_from_petab_problem(
            petab_problem, amici_model, edatas, method
        )

    def get_groups_for_xs(self, inner_parameter_type: str) -> List[int]:
        """Get unique list of ``OptimalScalingParameter.group`` values."""
        groups = [x.group for x in self.get_xs_for_type(inner_parameter_type)]
        return list(set(groups))

    def get_xs_for_group(self, group: int) -> List[OptimalScalingParameter]:
        r"""Get ``OptimalScalingParameter``\s that belong to the given group."""
        return [x for x in self.xs.values() if x.group == group]

    def get_free_xs_for_group(
        self, group: int
    ) -> List[OptimalScalingParameter]:
        r"""Get ``OptimalScalingParameter``\s that are free and belong to the given group."""
        return [
            x
            for x in self.xs.values()
            if x.group == group and x.estimate is True
        ]

    def get_fixed_xs_for_group(
        self, group: int
    ) -> List[OptimalScalingParameter]:
        r"""Get ``OptimalScalingParameter``\s that are fixed and belong to the given group."""
        return [
            x
            for x in self.xs.values()
            if x.group == group and x.estimate is False
        ]

    def get_cat_ub_parameters_for_group(
        self, group: int
    ) -> List[OptimalScalingParameter]:
        r"""Get ``OptimalScalingParameter``\s that are category upper boundaries and belong to the given group."""
        return [
            x
            for x in self.xs.values()
            if x.group == group and x.inner_parameter_id[:6] == CAT_UB
        ]

    def get_cat_lb_parameters_for_group(
        self, group: int
    ) -> List[OptimalScalingParameter]:
        r"""Get ``OptimalScalingParameter``\s that are category lower boundaries and belong to the given group."""
        return [
            x
            for x in self.xs.values()
            if x.group == group and x.inner_parameter_id[:6] == CAT_LB
        ]

    def initialize_c(self, group: int) -> np.ndarray:
        """Initialize the constraints matrix for the group.

        The structure of the constraints matrix is the following: Each row C_i of the matrix C
        represents one optimization constraint as C_i * xi + d(theta, sim) <= 0, where xi is the
        vector of inner paramters (surrogate data, lower bounds, upper bounds)^T, and d is the
        vector of minimal category interval ranges and gaps.

        First `self.groups[group][NUM_DATAPOINTS]` rows constrain the surrogate data to stay
        larger than lower category bounds.
        Then another `self.groups[group][NUM_DATAPOINTS]` rows constrain the surrogate data to
        stay smaller than upper category bounds.
        Then `self.groups[group][NUM_CATEGORIES]` rows constrain the ordering of the categories.
        And lastly, the remaining `self.groups[group][NUM_CATEGORIES]` constrain the lower
        bound to be smaller than the upper bound for each category.
        """
        constr = np.zeros(
            [
                self.groups[group][NUM_CONSTR_FULL],
                self.groups[group][NUM_INNER_PARAMS],
            ]
        )
        data_idx = 0

        # Iterate over categories.
        for cat_idx, category in enumerate(
            self.get_cat_ub_parameters_for_group(group)
        ):
            num_data_in_cat = int(
                np.sum(
                    [
                        np.sum(category.ixs[idx])
                        for idx in range(len(category.ixs))
                    ]
                )
            )

            # Constrain the surrogate data of this category to stay within it.
            for _ in range(num_data_in_cat):
                # lb - y_surr <= 0
                constr[data_idx, data_idx] = -1
                constr[
                    data_idx, cat_idx + self.groups[group][NUM_DATAPOINTS]
                ] = 1

                # y_surr - ub <= 0
                constr[
                    data_idx + self.groups[group][NUM_DATAPOINTS], data_idx
                ] = 1
                constr[
                    data_idx + self.groups[group][NUM_DATAPOINTS],
                    cat_idx
                    + self.groups[group][NUM_DATAPOINTS]
                    + self.groups[group][NUM_CATEGORIES],
                ] = -1
                data_idx += 1

            # Constrain the ordering wrt. neighbouring categories, i.e. ub_i - lb_{i+1} <= 0.
            if cat_idx == 0:
                constr[
                    2 * self.groups[group][NUM_DATAPOINTS] + cat_idx,
                    self.groups[group][NUM_DATAPOINTS] + cat_idx,
                ] = -1
            else:
                constr[
                    2 * self.groups[group][NUM_DATAPOINTS] + cat_idx,
                    self.groups[group][NUM_DATAPOINTS] + cat_idx,
                ] = -1
                constr[
                    2 * self.groups[group][NUM_DATAPOINTS] + cat_idx,
                    self.groups[group][NUM_DATAPOINTS]
                    + self.groups[group][NUM_CATEGORIES]
                    + cat_idx
                    - 1,
                ] = 1

            # Constrain upper bound to be larger than lower bound, i.e. lb_i - ub_i <= 0.
            constr[
                2 * self.groups[group][NUM_DATAPOINTS]
                + self.groups[group][NUM_CATEGORIES]  # - 1
                + cat_idx,
                self.groups[group][LB_INDICES][cat_idx],
            ] = 1
            constr[
                2 * self.groups[group][NUM_DATAPOINTS]
                + self.groups[group][NUM_CATEGORIES]  # - 1
                + cat_idx,
                self.groups[group][UB_INDICES][cat_idx],
            ] = -1

        return constr

    def initialize_w(self, group: int) -> np.ndarray:
        """Initialize the weight matrix for the group."""
        weights = np.diag(
            np.block(
                [
                    np.ones(self.groups[group][NUM_DATAPOINTS]),
                    np.zeros(2 * self.groups[group][NUM_CATEGORIES]),
                ]
            )
        )
        return weights

    def get_w(self, group: int, y_sim_all: np.ndarray) -> np.ndarray:
        """Return the weight matrix of the group."""
        weights = np.diag(
            np.block(
                [
                    np.ones(self.groups[group][NUM_DATAPOINTS])
                    / (np.sum(np.abs(y_sim_all)) + 1e-8),
                    np.zeros(2 * self.groups[group][NUM_CATEGORIES]),
                ]
            )
        )
        return weights

    def get_wdot(
        self, group: int, y_sim_all: np.ndarray, sy_all: np.ndarray
    ) -> np.ndarray:
        """Return the derivative of the weight matrix of a group with respect to an outer parameter."""
        w_dot = np.diag(
            np.block(
                [
                    np.ones(self.groups[group][NUM_DATAPOINTS])
                    * (
                        -1
                        * np.sum(sy_all)
                        / ((np.sum(np.abs(y_sim_all)) + 1e-8) ** 2)
                    ),
                    np.zeros(2 * self.groups[group][NUM_CATEGORIES]),
                ]
            )
        )
        return w_dot

    def get_d(
        self,
        group,
        xs: List[OptimalScalingParameter],
        y_sim_all: np.ndarray,
        eps: float,
    ) -> np.ndarray:
        """Return vector of minimal gaps and ranges."""
        max_simulation = np.nanmax(y_sim_all)

        interval_range = max_simulation / (2 * len(xs) + 1)
        interval_gap = max_simulation / (4 * (len(xs) - 1) + 1)

        d = np.zeros(self.groups[group][NUM_CONSTR_FULL])

        d[
            2 * self.groups[group][NUM_DATAPOINTS]
            + 1 : 2 * self.groups[group][NUM_DATAPOINTS]
            + self.groups[group][NUM_CATEGORIES]
        ] = (interval_gap + eps)

        d[
            2 * self.groups[group][NUM_DATAPOINTS]
            + self.groups[group][NUM_CATEGORIES] :
        ] = interval_range
        return d

    def get_dd_dtheta(
        self,
        group: int,
        xs: List[OptimalScalingParameter],
        y_sim_all: np.ndarray,
        sy_all: np.ndarray,
    ) -> np.ndarray:
        """Return the derivative of vector of minimal gaps and ranges with respect to an outer parameter."""
        max_sim_idx = np.argmax(y_sim_all)
        max_sy = sy_all[max_sim_idx]
        dd_dtheta = np.zeros(self.groups[group][NUM_CONSTR_FULL])

        dinterval_range_dtheta = max_sy / (2 * len(xs) + 1)
        dinterval_gap_dtheta = max_sy / (4 * (len(xs) - 1) + 1)

        dd_dtheta[
            2 * self.groups[group][NUM_DATAPOINTS]
            + 1 : 2 * self.groups[group][NUM_DATAPOINTS]
            + self.groups[group][NUM_CATEGORIES]
        ] = dinterval_gap_dtheta

        dd_dtheta[
            2 * self.groups[group][NUM_DATAPOINTS]
            + self.groups[group][NUM_CATEGORIES] :
        ] = dinterval_range_dtheta

        return dd_dtheta

    def get_censored_group_quantitative_ixs(
        self, xs: List[OptimalScalingParameter]
    ) -> List[np.ndarray]:
        r"""Return a list of boolean masks indicating which data points are quantitative.

        For a given group with censored data, return a list of boolean masks indicating
        which data points are not censored, and therefore quantitative.

        Parameters
        ----------
        xs:
            List of ``OptimalScalingParameter``\s of a group with censored data.

        Returns
        -------
        quantitative_ixs:
            List of boolean masks indicating which data points are quantitative.
        """
        # Initialize boolean masks with False and find corresponding observable index.
        quantitative_ixs = [np.full(ixs_i.shape, False) for ixs_i in xs[0].ixs]
        observable_index = xs[0].group - 1

        # Set to True all datapoints of the corresponding observable.
        if np.ndim(quantitative_ixs) == 2:
            quantitative_ixs = [
                np.full(ixs_i.shape, True) for ixs_i in xs[0].ixs
            ]
        else:
            for quantitative_ixs_i in quantitative_ixs:
                quantitative_ixs_i[:, observable_index] = True

        # Set to False for all censored datapoints.
        for x in xs:
            for ixs_i, quantitative_ixs_i in zip(x.ixs, quantitative_ixs):
                quantitative_ixs_i[ixs_i] = False

        return quantitative_ixs

    def get_inner_parameter_dictionary(self) -> Dict:
        """Return a dictionary with inner parameter ids and their values."""
        inner_par_dict = {}
        for x_id, x in self.xs.items():
            inner_par_dict[x_id] = x.value
        return inner_par_dict


def optimal_scaling_inner_problem_from_petab_problem(
    petab_problem: petab.Problem,
    amici_model: 'amici.Model',
    edatas: List['amici.ExpData'],
    method: str,
):
    """Construct the inner problem from the `petab_problem`."""
    # inner parameters
    inner_parameters = optimal_scaling_inner_parameters_from_measurement_df(
        petab_problem.measurement_df, method, amici_model
    )

    # used indices for all measurement specific parameters
    ixs = optimal_scaling_ixs_for_measurement_specific_parameters(
        petab_problem, amici_model, inner_parameters
    )

    # transform experimental data
    edatas = [
        amici.numpy.ExpDataView(edata)['observedData'] for edata in edatas
    ]

    # matrixify
    ix_matrices = ix_matrices_from_arrays(ixs, edatas)

    # assign matrices
    for par in inner_parameters:
        par.ixs = ix_matrices[par.inner_parameter_id]

    return OptimalScalingProblem(
        inner_parameters,
        edatas,
        method,
    )


def optimal_scaling_inner_parameters_from_measurement_df(
    df: pd.DataFrame,
    method: str,
    amici_model: 'amici.Model',
) -> List[OptimalScalingParameter]:
    """Create list of inner free parameters from PEtab measurement table dependent on the method provided."""
    df = df.reset_index()

    observable_ids = amici_model.getObservableIds()

    estimate = get_estimate_for_method(method)
    par_types = [CAT_LB, CAT_UB]

    inner_parameters = []
    lb, ub = INNER_PARAMETER_BOUNDS[
        InnerParameterType.OPTIMAL_SCALING
    ].values()

    for observable_idx, observable_id in enumerate(observable_ids):
        group = observable_idx + 1

        observable_df = df[df[OBSERVABLE_ID] == observable_id]

        if all(observable_df[MEASUREMENT_TYPE] == ORDINAL):
            # Add optimal scaling parameters for ordinal measurements.
            for par_type, par_estimate in zip(par_types, estimate):
                for _, row in observable_df.iterrows():
                    par_id = f'{par_type}_{observable_id}_{row[MEASUREMENT_TYPE]}_{int(row[MEASUREMENT_CATEGORY])}'

                    # Create only one set of bound parameters per category of a group.
                    if par_id not in [
                        inner_par.inner_parameter_id
                        for inner_par in inner_parameters
                    ]:
                        inner_parameters.append(
                            OptimalScalingParameter(
                                inner_parameter_id=par_id,
                                inner_parameter_type=InnerParameterType.OPTIMAL_SCALING,
                                scale=LIN,
                                lb=lb,
                                ub=ub,
                                observable_id=observable_id,
                                category=int(row[MEASUREMENT_CATEGORY]),
                                group=group,
                                estimate=par_estimate,
                            )
                        )
        elif any(observable_df[MEASUREMENT_TYPE].isin(CENSORING_TYPES)):
            # Get df with only censored measurements.
            censored_df = observable_df.loc[
                observable_df[MEASUREMENT_TYPE].isin(CENSORING_TYPES)
            ]
            # Check for unique values in the CENSORING_BOUNDS column and
            # order them with resect to the first float value in the string.
            unique_censoring_bounds = sorted(
                censored_df[CENSORING_BOUNDS].unique(),
                key=lambda x: float(str(x).split(PARAMETER_SEPARATOR)[0]),
            )
            for par_type in par_types:
                for _, row in censored_df.iterrows():
                    category = int(
                        unique_censoring_bounds.index(row[CENSORING_BOUNDS])
                        + 1
                    )
                    par_id = f'{par_type}_{observable_id}_{row[MEASUREMENT_TYPE]}_{category}'
                    # Create only one set of bound parameters per category of a group.
                    if par_id not in [
                        inner_par.inner_parameter_id
                        for inner_par in inner_parameters
                    ]:
                        inner_parameters.append(
                            OptimalScalingParameter(
                                inner_parameter_id=par_id,
                                inner_parameter_type=InnerParameterType.OPTIMAL_SCALING,
                                scale=LIN,
                                lb=lb,
                                ub=ub,
                                observable_id=observable_id,
                                category=category,
                                group=group,
                                estimate=False,
                                censoring_type=row[MEASUREMENT_TYPE],
                            )
                        )
                        _add_value_to_censored_bound_parameter(
                            inner_parameters[-1], row, par_type
                        )

    inner_parameters.sort(key=lambda x: (x.group, x.category))

    return inner_parameters


def get_estimate_for_method(method: str) -> Tuple[bool, bool]:
    """Return which inner parameters to estimate dependent on the method provided."""
    estimate_ub = True
    estimate_lb = False

    if method == STANDARD:
        estimate_lb = True

    return estimate_lb, estimate_ub


def optimal_scaling_ixs_for_measurement_specific_parameters(
    petab_problem: 'petab.Problem',
    amici_model: 'amici.Model',
    inner_parameters: List[OptimalScalingParameter],
) -> Dict[str, List[Tuple[int, int, int]]]:
    """Create mapping of parameters to measurements.

    Returns
    -------
    A dictionary mapping parameter ID to a list of
    `(condition index, time index, observable index)` tuples in which this
    output parameter is used. For each condition, the time index refers to
    a sorted list of non-unique time points for which there are measurements.
    """
    ixs_for_par = {}
    observable_ids = amici_model.getObservableIds()

    simulation_conditions = (
        petab_problem.get_simulation_conditions_from_measurement_df()
    )
    measurement_df = petab_problem.measurement_df
    # Get unique censoring bounds for each observable.
    unique_censoring_bounds_per_observable = {}
    if CENSORING_BOUNDS in measurement_df.columns:
        for observable_id in observable_ids:
            observable_df = measurement_df[
                measurement_df[OBSERVABLE_ID] == observable_id
            ]
            censored_observable_df = observable_df.loc[
                observable_df[MEASUREMENT_TYPE].isin(CENSORING_TYPES)
            ]
            unique_censoring_bounds_per_observable[observable_id] = sorted(
                censored_observable_df[CENSORING_BOUNDS].unique(),
                key=lambda x: float(str(x).split(PARAMETER_SEPARATOR)[0]),
            )

    for condition_ix, condition in simulation_conditions.iterrows():
        # measurement table for current condition
        df_for_condition = petab.get_rows_for_condition(
            measurement_df=measurement_df, condition=condition
        )

        # unique sorted list of timepoints
        timepoints = sorted(df_for_condition[TIME].unique().astype(float))
        # non-unique sorted list of timepoints
        timepoints_w_reps = _get_timepoints_with_replicates(
            measurement_df=df_for_condition
        )

        for time in timepoints:
            # subselect measurements for time `time`
            df_for_time = df_for_condition[df_for_condition[TIME] == time]
            time_ix_0 = timepoints_w_reps.index(time)

            # remember used time indices for each observable
            time_ix_for_obs_ix = {}

            # iterate over measurements
            for _, measurement in df_for_time.iterrows():
                # extract observable index
                observable_ix = observable_ids.index(
                    measurement[OBSERVABLE_ID]
                )

                # as the time indices have to account for replicates, we need
                #  to track which time indices have already been assigned for
                #  the current observable
                if observable_ix in time_ix_for_obs_ix:
                    # a replicate
                    time_ix_for_obs_ix[observable_ix] += 1
                else:
                    # the first measurement for this `(observable, timepoint)`
                    time_ix_for_obs_ix[observable_ix] = time_ix_0
                time_w_reps_ix = time_ix_for_obs_ix[observable_ix]

                inner_par_ids_for_meas = get_inner_par_ids_for_measurement(
                    measurement,
                    inner_parameters,
                    unique_censoring_bounds_per_observable,
                )

                # try to insert if hierarchical parameter
                if inner_par_ids_for_meas:
                    for override in inner_par_ids_for_meas:
                        ixs_for_par.setdefault(override, []).append(
                            (condition_ix, time_w_reps_ix, observable_ix)
                        )
    return ixs_for_par


def get_inner_par_ids_for_measurement(
    measurement: Dict,
    inner_parameters: List[OptimalScalingParameter],
    unique_censoring_bounds_per_observable: Dict[str, List[float]],
):
    """Return inner parameter ids of parameters which are related to the measurement."""
    if measurement[MEASUREMENT_TYPE] == ORDINAL:
        return [
            inner_par.inner_parameter_id
            for inner_par in inner_parameters
            if inner_par.category == measurement[MEASUREMENT_CATEGORY]
            and inner_par.observable_id == measurement[OBSERVABLE_ID]
        ]
    elif measurement[MEASUREMENT_TYPE] in [
        LEFT_CENSORED,
        INTERVAL_CENSORED,
        RIGHT_CENSORED,
    ]:
        unique_censoring_bounds = unique_censoring_bounds_per_observable[
            measurement[OBSERVABLE_ID]
        ]
        measurement_category = int(
            unique_censoring_bounds.index(measurement[CENSORING_BOUNDS]) + 1
        )
        return [
            inner_par.inner_parameter_id
            for inner_par in inner_parameters
            if inner_par.observable_id == measurement[OBSERVABLE_ID]
            and inner_par.category == measurement_category
        ]


def _add_value_to_censored_bound_parameter(
    inner_parameter: OptimalScalingParameter,
    row: pd.Series,
    par_type: str,
) -> None:
    if row[MEASUREMENT_TYPE] == LEFT_CENSORED and par_type == CAT_LB:
        inner_parameter.value = 0.0
    elif row[MEASUREMENT_TYPE] == LEFT_CENSORED and par_type == CAT_UB:
        inner_parameter.value = float(row[CENSORING_BOUNDS])

    elif row[MEASUREMENT_TYPE] == RIGHT_CENSORED and par_type == CAT_LB:
        inner_parameter.value = float(row[CENSORING_BOUNDS])
    elif row[MEASUREMENT_TYPE] == RIGHT_CENSORED and par_type == CAT_UB:
        inner_parameter.value = np.inf

    elif row[MEASUREMENT_TYPE] == INTERVAL_CENSORED and par_type == CAT_LB:
        inner_parameter.value = float(
            row[CENSORING_BOUNDS].split(PARAMETER_SEPARATOR)[0]
        )
    elif row[MEASUREMENT_TYPE] == INTERVAL_CENSORED and par_type == CAT_UB:
        inner_parameter.value = float(
            row[CENSORING_BOUNDS].split(PARAMETER_SEPARATOR)[1]
        )



"""Self-adaptive cooperative enhanced scatter search (SACESS)."""
import itertools
import logging
import os
import time
from multiprocessing import Manager, Process
from multiprocessing.managers import SyncManager
from typing import Any, Dict, List, Optional, Tuple

import numpy as np

import pypesto

from ...startpoint import StartpointMethod
from ...store.read_from_hdf5 import read_result
from ...store.save_to_hdf5 import write_result
from ..optimize import Problem
from .ess import ESSExitFlag, ESSOptimizer
from .function_evaluator import FunctionEvaluatorMP, FunctionEvaluatorMT
from .refset import RefSet

__all__ = ["SacessOptimizer", "get_default_ess_options"]

logger = logging.getLogger(__name__)


class SacessOptimizer:
    """SACESS optimizer.

    A shared-memory-based implementation of the SaCeSS algorithm presented in
    [PenasGon2017]_. Multiple processes (`workers`) run consecutive ESSs in
    parallel. After each ESS run, depending on the outcome, there is a chance
    of exchanging good parameters, and changing ESS hyperparameters to those of
    the most promising worker.

    .. [PenasGon2017] 'Parameter estimation in large-scale systems biology models:
       a parallel and self-adaptive cooperative strategy', David R. Penas,
       Patricia González, Jose A. Egea, Ramón Doallo and Julio R. Banga,
       BMC Bioinformatics 2017, 18, 52. https://doi.org/10.1186/s12859-016-1452-4
    """

    def __init__(
        self,
        num_workers: Optional[int] = None,
        ess_init_args: Optional[List[Dict[str, Any]]] = None,
        max_walltime_s: float = np.inf,
        sacess_loglevel: int = logging.INFO,
        ess_loglevel: int = logging.WARNING,
    ):
        """Construct.

        Parameters
        ----------
        ess_init_args:
            List of argument dictionaries passed to
            :func:`ESSOptimizer.__init__`. Each entry corresponds to one worker
            process. I.e., the length of this list is the number of ESSs.
            Ideally, this list contains some more conservative and some more
            aggressive configurations.
            Resource limits such as ``max_eval`` apply to a single CESS
            iteration, not to the full search.
            Mutually exclusive with ``num_workers``.
        num_workers:
            Number of workers to be used. If this argument is given,
            (different) default ESS settings will be used for each worker.
            Mutually exclusive with ``ess_init_args``.
        max_walltime_s:
            Maximum walltime in seconds. Will only be checked between local
            optimizations and other simulations, and thus, may be exceeded by
            the duration of a local search. Defaults to no limit.
        ess_loglevel:
            Loglevel for ESS runs.
        sacess_loglevel:
            Loglevel for SACESS runs.
        """
        if (num_workers is None and ess_init_args is None) or (
            num_workers is not None and ess_init_args is not None
        ):
            raise ValueError(
                "Exactly one of `num_workers` or `ess_init_args` "
                "has to be provided."
            )

        self.num_workers = num_workers or len(ess_init_args)
        if self.num_workers < 2:
            raise ValueError(
                f"{self.__class__.__name__} needs at least 2 workers."
            )
        self.ess_init_args = ess_init_args
        self.max_walltime_s = max_walltime_s
        self.exit_flag = ESSExitFlag.DID_NOT_RUN
        self.ess_loglevel = ess_loglevel
        self.sacess_loglevel = sacess_loglevel
        logger.setLevel(self.sacess_loglevel)

    def minimize(
        self,
        problem: Problem,
        startpoint_method: StartpointMethod,
    ):
        """Solve the given optimization problem."""
        start_time = time.time()
        logger.debug(
            f"Running sacess with {self.num_workers} "
            f"workers: {self.ess_init_args}"
        )
        ess_init_args = self.ess_init_args or get_default_ess_options(
            num_workers=self.num_workers, dim=problem.dim
        )

        # shared memory manager to handle shared state
        # (simulates the sacess manager process)
        with Manager() as shmem_manager:
            sacess_manager = SacessManager(
                shmem_manager=shmem_manager,
                ess_options=ess_init_args,
                dim=problem.dim,
            )
            # create workers
            workers = [
                SacessWorker(
                    manager=sacess_manager,
                    ess_kwargs=ess_kwargs,
                    worker_idx=i,
                    max_walltime_s=self.max_walltime_s,
                    ess_loglevel=self.ess_loglevel,
                )
                for i, ess_kwargs in enumerate(ess_init_args)
            ]
            # launch worker processes
            worker_processes = [
                Process(
                    name=f"sacess-worker-{i:02d}",
                    target=_run_worker,
                    args=(
                        worker,
                        problem,
                        startpoint_method,
                        self.sacess_loglevel,
                    ),
                )
                for i, worker in enumerate(workers)
            ]
            for p in worker_processes:
                p.start()
            # wait for finish
            for p in worker_processes:
                p.join()

            walltime = time.time() - start_time
            logger.info(
                f"sacess stopping after {walltime:3g}s with global best "
                f"{sacess_manager.get_best_solution()[1]}."
            )

        return self._create_result(problem)

    def _create_result(self, problem: Problem) -> pypesto.Result:
        """Create result object.

        Creates an overall Result object from the results saved by the workers.
        """
        # gather results from workers and delete temporary result files
        result = None
        for worker_idx in range(self.num_workers):
            tmp_result_filename = SacessWorker.get_temp_result_filename(
                worker_idx
            )
            tmp_result = read_result(
                tmp_result_filename, problem=False, optimize=True
            )
            os.remove(tmp_result_filename)
            if result is None:
                result = tmp_result
            else:
                result.optimize_result.append(
                    tmp_result.optimize_result,
                    sort=False,
                    prefix=f"{worker_idx}-",
                )
        result.optimize_result.sort()

        result.problem = problem

        return result


class SacessManager:
    """The Sacess manager.

    Manages shared memory of a SACESS run. Loosely corresponds to the manager
    process in [PenasGon2017]_.

    Attributes
    ----------
    _num_workers: Number of workers
    _ess_options: ESS options for each worker
    _best_known_fx: Best objective value encountered so far
    _best_known_x: Parameters corresponding to ``_best_known_fx``
    _worker_scores: Performance score of the different workers (the higher, the
        more promising the respective worker is considered)
    _worker_comms: Number of communications received from the individual
        workers
    _rejections: Number of rejected solutions received from workers since last
        adaptation of ``_rejection_threshold``.
    _rejection_threshold: Threshold for relative objective improvements that
        incoming solutions have to pass to be accepted
    _lock: Lock for accessing shared state.
    _logger: A logger instance
    """

    def __init__(
        self,
        shmem_manager: SyncManager,
        ess_options: List[Dict[str, Any]],
        dim: int,
    ):
        self._num_workers = len(ess_options)
        self._ess_options = [shmem_manager.dict(o) for o in ess_options]
        self._best_known_fx = shmem_manager.Value("d", np.inf)
        self._best_known_x = shmem_manager.Array("d", [np.nan] * dim)
        self._rejections = shmem_manager.Value("i", 0)
        # initial value from [PenasGon2017]_ p.9
        self._rejection_threshold = shmem_manager.Value("d", 0.1)
        # scores of the workers, ordered by worker-index
        # initial score is the worker index
        self._worker_scores = shmem_manager.Array(
            "d", range(self._num_workers)
        )
        self._worker_comms = shmem_manager.Array("i", [0] * self._num_workers)
        self._lock = shmem_manager.RLock()
        self._logger = logging.getLogger()

    def get_best_solution(self) -> Tuple[np.array, float]:
        """Get the best objective value and corresponding parameters."""
        with self._lock:
            return np.array(self._best_known_x), self._best_known_fx.value

    def reconfigure_worker(self, worker_idx: int) -> Dict:
        """Reconfigure the given worker.

        Updates the ESS options for the given worker to those of the worker at
        the top of the scoreboard and returns those settings.
        """
        with self._lock:
            leader_options = self._ess_options[np.argmax(self._worker_scores)]
            self._ess_options[worker_idx] = leader_options
            return leader_options

    def submit_solution(
        self,
        x: np.array,
        fx: float,
        sender_idx: int,
        elapsed_time_s: float,
    ):
        """Submit a solution.

        To be called by a worker.

        Parameters
        ----------
        x: Model parameters
        fx: Objective value corresponding to ``x``
        sender_idx: Index of the worker submitting the results.
        elapsed_time_s: Elapsed time since the beginning of the sacess run.
        """
        with self._lock:
            # cooperation step
            # solution improves best value by at least a factor of ...
            if (
                # initially _best_known_fx is NaN
                (
                    np.isfinite(fx)
                    and not np.isfinite(self._best_known_fx.value)
                )
                # avoid division by 0. just accept any improvement if best
                # known value is 0.
                or (self._best_known_fx.value == 0 and fx < 0)
                or (
                    fx < self._best_known_fx.value
                    and abs(
                        (self._best_known_fx.value - fx)
                        / self._best_known_fx.value
                    )
                    > self._rejection_threshold.value
                )
            ):
                # accept solution
                self._logger.debug(
                    f"Accepted solution from worker {sender_idx}: {fx}."
                )
                # accept
                self._best_known_fx.value = fx
                self._best_known_x.value = x
                self._worker_comms[sender_idx] += 1
                self._worker_scores[sender_idx] = (
                    self._worker_comms[sender_idx] * elapsed_time_s
                )
            else:
                # reject solution
                self._rejections.value += 1
                self._logger.debug(
                    f"Rejected solution from worker {sender_idx} "
                    f"rel change: {abs((self._best_known_fx.value - fx) / self._best_known_fx.value)} "
                    f" < {self._rejection_threshold.value} "
                    f"(total rejections: {self._rejections.value})."
                )
                # adapt acceptance threshold if too many solutions have been
                #  rejected
                if self._rejections.value > self._num_workers:
                    self._rejection_threshold.value /= 2
                    self._logger.debug(
                        "Lowered acceptance threshold to "
                        f"{self._rejection_threshold.value}."
                    )
                    self._rejections.value = 0


class SacessWorker:
    """A SACESS worker.

    Repeatedly runs ESSs and exchanges information with a SacessManager.
    Corresponds to a worker process in [PenasGon2017]_.

    Attributes
    ----------
    _manager: The sacess manager this worker is working for.
    _worker_idx: Index of this worker.
    _best_known_fx: Best objective value known to this worker (obtained on its
        own or received from the manager).
    _n_received_solutions: Number of solutions received by this worker since
        the last one was sent to the manager.
    _neval: Number of objective evaluations since the last solution was sent
        to the manager.
    _ess_kwargs: ESSOptimizer options for this worker (may get updated during
        the self-adaptive step).
    _acceptance_threshold: Minimum relative improvement of the objective
        compared to the best known value to be eligible for submission to the
        Manager.
    _n_sent_solutions: Number of solutions sent to the Manager.
    _max_walltime_s: Walltime limit.
    _logger: A Logger instance.
    _ess_loglevel: Logging level for ESS runs
    """

    def __init__(
        self,
        manager: SacessManager,
        ess_kwargs: Dict[str, Any],
        worker_idx: int,
        max_walltime_s: float = np.inf,
        ess_loglevel: int = logging.WARNING,
    ):
        self._manager = manager
        self._worker_idx = worker_idx
        self._best_known_fx = np.inf
        self._n_received_solutions = 0
        self._neval = 0
        self._ess_kwargs = ess_kwargs
        self._acceptance_threshold = 0.005
        self._n_sent_solutions = 0
        self._max_walltime_s = max_walltime_s
        self._start_time = None
        self._logger = logging.getLogger(str(os.getpid()))
        # Set the manager logger to one created within the current process
        self._manager._logger = self._logger
        self._ess_loglevel = ess_loglevel

    def run(
        self,
        problem: Problem,
        startpoint_method: StartpointMethod,
    ):
        """Start the worker."""
        self._logger.debug(
            f"#{self._worker_idx} starting " f"({self._ess_kwargs})."
        )

        if n_procs := self._ess_kwargs.get('n_procs'):
            evaluator = FunctionEvaluatorMP(
                problem=problem,
                startpoint_method=startpoint_method,
                n_procs=n_procs,
            )
        else:
            evaluator = FunctionEvaluatorMT(
                problem=problem,
                startpoint_method=startpoint_method,
                n_threads=self._ess_kwargs.get('n_threads', 1),
            )
        self._start_time = time.time()
        # create refset from ndiverse
        refset = RefSet(
            dim=self._ess_kwargs['dim_refset'], evaluator=evaluator
        )
        refset.initialize_random(
            n_diverse=max(
                self._ess_kwargs.get('n_diverse', 10 * problem.dim), refset.dim
            )
        )
        i_iter = 0
        ess_results = pypesto.Result(problem=problem)

        while self._keep_going():
            # run standard ESS
            ess, cur_ess_results = self._run_ess(
                refset=refset,
            )

            # drop all but the 50 best results
            ess_results.optimize_result.append(
                cur_ess_results.optimize_result,
                prefix=f"{self._worker_idx}_{i_iter}_",
            )
            ess_results.optimize_result.list = (
                ess_results.optimize_result.list[:50]
            )
            write_result(
                ess_results,
                self.get_temp_result_filename(self._worker_idx),
                overwrite=True,
                optimize=True,
            )
            # check if the best solution of the last local ESS is sufficiently
            # better than the sacess-wide best solution
            self.maybe_update_best(ess.x_best, ess.fx_best)
            self._best_known_fx = min(ess.fx_best, self._best_known_fx)

            # cooperation step
            # try to obtain a new solution from manager
            recv_x, recv_fx = self._manager.get_best_solution()
            self._logger.log(
                logging.DEBUG - 1,
                f"Worker {self._worker_idx} received solution {recv_fx} "
                f"(known best: {self._best_known_fx}).",
            )
            if recv_fx < self._best_known_fx or (
                not np.isfinite(self._best_known_fx) and np.isfinite(recv_x)
            ):
                self._logger.debug(
                    f"Worker {self._worker_idx} received better solution."
                )
                self._best_known_fx = recv_fx
                self._n_received_solutions += 1
                self.replace_solution(refset, x=recv_x, fx=recv_fx)

            # Adaptive step
            # Update ESS settings if we received way more solutions than we
            # sent
            # Magic numbers from [PenasGon2017]_ algorithm 5
            if (
                self._n_received_solutions > 10 * self._n_sent_solutions + 20
                and self._neval > problem.dim * 5000
            ):
                self._ess_kwargs = self._manager.reconfigure_worker(
                    self._worker_idx
                )
                self._logger.debug(
                    f"Updated settings on worker {self._worker_idx} to "
                    f"{self._ess_kwargs}"
                )

            self._logger.info(
                f"sacess worker {self._worker_idx} iteration {i_iter} "
                f"(best: {self._best_known_fx})."
            )

            i_iter += 1

    def _run_ess(
        self,
        refset: RefSet,
    ) -> Tuple[ESSOptimizer, pypesto.Result]:
        """Run ESS."""
        ess_kwargs = self._ess_kwargs.copy()
        # account for sacess walltime limit
        ess_kwargs['max_walltime_s'] = min(
            ess_kwargs.get('max_walltime_s', np.inf),
            self._max_walltime_s - (time.time() - self._start_time),
        )

        ess = ESSOptimizer(**ess_kwargs)
        ess.logger.setLevel(self._ess_loglevel)

        cur_ess_results = ess.minimize(
            refset=refset,
        )
        self._logger.debug(
            f"#{self._worker_idx}: ESS finished with best "
            f"f(x)={ess.fx_best}"
        )
        return ess, cur_ess_results

    def maybe_update_best(self, x: np.array, fx: float):
        """Maybe update the best known solution and send it to the manager."""
        self._logger.debug(
            f"Worker {self._worker_idx} maybe sending solution {fx}. "
            f"best known: {self._best_known_fx}, "
            f"rel change: {(self._best_known_fx - fx) / fx}, "
            f"threshold: {self._acceptance_threshold}"
        )

        # solution improves best value by at least a factor of ...
        if (
            (np.isfinite(fx) and not np.isfinite(self._best_known_fx))
            or (self._best_known_fx == 0 and fx < 0)
            or (
                fx < self._best_known_fx
                and abs((self._best_known_fx - fx) / fx)
                > self._acceptance_threshold
            )
        ):
            self._logger.debug(
                f"Worker {self._worker_idx} sending solution {fx}."
            )
            self._n_sent_solutions += 1
            self._best_known_fx = fx

            # send best_known_fx to manager
            self._neval = 0
            self._n_received_solutions = 0
            elapsed_time = time.time() - self._start_time
            self._manager.submit_solution(
                x=x,
                fx=fx,
                sender_idx=self._worker_idx,
                elapsed_time_s=elapsed_time,
            )

    @staticmethod
    def replace_solution(refset: RefSet, x: np.array, fx: float):
        """Replace the global refset member by the given solution."""
        # [PenasGon2017]_ page 8, top
        if "cooperative_solution" not in refset.attributes:
            label = np.zeros(shape=refset.dim)
            # on first call, mark the worst solution as "cooperative solution"
            label[np.argmax(refset.fx)] = 1
            refset.add_attribute("cooperative_solution", label)

        # replace the cooperative solution
        refset.update(
            i=np.argwhere(refset.attributes["cooperative_solution"]),
            x=x,
            fx=fx,
        )

    def _keep_going(self):
        """Check exit criteria.

        Returns
        -------
        ``True`` if none of the exit criteria is met, ``False`` otherwise.
        """
        # elapsed time
        if time.time() - self._start_time >= self._max_walltime_s:
            self.exit_flag = ESSExitFlag.MAX_TIME
            self._logger.debug(
                f"Max walltime ({self._max_walltime_s}s) exceeded."
            )
            return False

        return True

    @staticmethod
    def get_temp_result_filename(worker_idx: int) -> str:
        return f"sacess-{worker_idx:02d}_tmp.h5"


def _run_worker(
    worker: SacessWorker,
    problem: Problem,
    startpoint_method: StartpointMethod,
    sacess_loglevel: int,
):
    """Run the given SACESS worker.

    Helper function as entrypoint for sacess worker processes.
    """
    # different random seeds per process
    np.random.seed((os.getpid() * int(time.time() * 1000)) % 2**32)

    worker._logger.setLevel(sacess_loglevel)

    return worker.run(problem=problem, startpoint_method=startpoint_method)


def get_default_ess_options(num_workers: int, dim: int) -> List[Dict]:
    """Get default ESS settings for (SA)CESS.

    Returns settings for ``num_workers`` parallel scatter searches, combining
    more aggressive and more conservative configurations.

    Setting appropriate values for ``n_threads`` and ``local_optimizer`` is
    left to the user. Defaults to single-threaded and no local optimizer.

    Parameters
    ----------
    num_workers: Number of configurations to return.
    dim: Problem dimension.
    """
    min_dimrefset = 3
    settings = [
        # settings for first worker
        {
            'dim_refset': 10 * dim,
            'balance': 0.5,
            'local_n2': 10,
        },
        # for the remaining workers, cycle through these settings
        # 1
        {
            'dim_refset': max(min_dimrefset, dim),
            'balance': 0.0,
            'local_n1': 1,
            'local_n2': 1,
        },
        # 2
        {
            'dim_refset': 3 * dim,
            'balance': 0.0,
            'local_n1': 4,
            'local_n2': 4,
        },
        # 3
        {
            'dim_refset': 5 * dim,
            'balance': 0.25,
            'local_n1': 10,
            'local_n2': 10,
        },
        # 4
        {
            'dim_refset': 10 * dim,
            'balance': 0.5,
            'local_n1': 20,
            'local_n2': 20,
        },
        # 5
        {
            'dim_refset': 15 * dim,
            'balance': 0.25,
            'local_n1': 100,
            'local_n2': 100,
        },
        # 6
        {
            'dim_refset': 12 * dim,
            'balance': 0.25,
            'local_n1': 50,
            'local_n2': 50,
        },
        # 7
        {
            'dim_refset': int(7.5 * dim),
            'balance': 0.25,
            'local_n1': 15,
            'local_n2': 15,
        },
        # 8
        {
            'dim_refset': 5 * dim,
            'balance': 0.25,
            'local_n1': 7,
            'local_n2': 7,
        },
        # 9
        {
            'dim_refset': max(min_dimrefset, 2 * dim),
            'balance': 0.0,
            'local_n1': 2,
            'local_n2': 2,
        },
        # 10
        {
            'dim_refset': max(min_dimrefset, int(0.5 * dim)),
            'balance': 0.0,
            'local_n1': 1,
            'local_n2': 1,
        },
    ]
    return [
        settings[0],
        *(itertools.islice(itertools.cycle(settings[1:]), num_workers - 1)),
    ]



"""Module for the InnerCalculatorCollector class.

In case of non-quantitative measurements, this class is used to collect
hierarchical inner calculators for each data type and merge their results.
"""
from __future__ import annotations

import copy
from typing import Dict, List, Sequence, Tuple, Union

import numpy as np

from ..C import (
    AMICI_SIGMAY,
    AMICI_SSIGMAY,
    AMICI_SSIGMAZ,
    AMICI_SY,
    AMICI_Y,
    CENSORED,
    FVAL,
    GRAD,
    HESS,
    MEASUREMENT_TYPE,
    METHOD,
    MODE_RES,
    NONLINEAR_MONOTONE,
    OPTIMAL_SCALING_OPTIONS,
    ORDINAL,
    RDATAS,
    RES,
    SPLINE_APPROXIMATION_OPTIONS,
    SPLINE_RATIO,
    SRES,
    X_INNER_OPT,
    ModeType,
)
from ..objective.amici.amici_calculator import AmiciCalculator
from ..objective.amici.amici_util import filter_return_dict, init_return_values

try:
    import amici
    import petab
    from amici.parameter_mapping import ParameterMapping
    from petab.C import OBSERVABLE_ID
except ImportError:
    petab = None
    ParameterMapping = None

from .optimal_scaling import (
    OptimalScalingAmiciCalculator,
    OptimalScalingInnerSolver,
    OptimalScalingProblem,
)
from .spline_approximation import (
    SplineAmiciCalculator,
    SplineInnerProblem,
    SplineInnerSolver,
)

AmiciModel = Union['amici.Model', 'amici.ModelPtr']
AmiciSolver = Union['amici.Solver', 'amici.SolverPtr']


class InnerCalculatorCollector(AmiciCalculator):
    """Class to collect inner calculators in case of non-quantitative data types.

    Upon import of a petab problem, the PEtab importer checks whether there are
    non-quantitative data types. If so, it creates an instance of this class
    instead of an AmiciCalculator. This class then collects the inner calculators
    for each data type and merges their results with the quantitative results.

    Parameters
    ----------
    data_types:
        List of non-quantitative data types in the problem.
    petab_problem:
        The PEtab problem.
    model:
        The AMICI model.
    edatas:
        The experimental data.
    inner_options:
        Options for the inner problems and solvers.
    """

    def __init__(
        self,
        data_types: List[str],
        petab_problem: 'petab.Problem',
        model: AmiciModel,
        edatas: List['amici.ExpData'],
        inner_options: Dict,
    ):
        super().__init__()
        self.validate_options(inner_options)

        self.data_types = data_types
        self.inner_calculators = []
        self.construct_inner_calculators(
            petab_problem, model, edatas, inner_options
        )

        self.quantitative_data_mask = self._get_quantitative_data_mask(
            petab_problem, model.getObservableIds(), edatas
        )

        self._known_least_squares_safe = False

    def initialize(self):
        """Initialize."""
        for calculator in self.inner_calculators:
            calculator.initialize()

    def construct_inner_calculators(
        self,
        petab_problem: 'petab.Problem',
        model: AmiciModel,
        edatas: List['amici.ExpData'],
        inner_options: Dict,
    ):
        """Construct inner calculators for each data type."""
        if ORDINAL in self.data_types or CENSORED in self.data_types:
            optimal_scaling_inner_options = {
                key: value
                for key, value in inner_options.items()
                if key in OPTIMAL_SCALING_OPTIONS
            }
            inner_problem_method = optimal_scaling_inner_options.get(
                METHOD, None
            )
            os_inner_problem = OptimalScalingProblem.from_petab_amici(
                petab_problem, model, edatas, inner_problem_method
            )
            os_inner_solver = OptimalScalingInnerSolver(
                options=optimal_scaling_inner_options
            )
            os_calculator = OptimalScalingAmiciCalculator(
                os_inner_problem, os_inner_solver
            )
            self.inner_calculators.append(os_calculator)

        if NONLINEAR_MONOTONE in self.data_types:
            spline_inner_options = {
                key: value
                for key, value in inner_options.items()
                if key in SPLINE_APPROXIMATION_OPTIONS
            }
            spline_ratio = spline_inner_options.pop(SPLINE_RATIO, None)
            spline_inner_problem = SplineInnerProblem.from_petab_amici(
                petab_problem, model, edatas, spline_ratio
            )
            spline_inner_solver = SplineInnerSolver(
                options=spline_inner_options
            )
            spline_calculator = SplineAmiciCalculator(
                spline_inner_problem, spline_inner_solver
            )
            self.inner_calculators.append(spline_calculator)
        # TODO relative data

        if set(self.data_types) - {ORDINAL, CENSORED, NONLINEAR_MONOTONE}:
            unsupported_data_types = set(self.data_types) - {
                ORDINAL,
                CENSORED,
                NONLINEAR_MONOTONE,
            }
            raise NotImplementedError(
                f"Data types {unsupported_data_types} are not supported."
            )

    def validate_options(self, inner_options: Dict):
        """Validate the inner options.

        Parameters
        ----------
        inner_options:
            Options for the inner problems and solvers.
        """
        for key in inner_options:
            if (
                key not in OPTIMAL_SCALING_OPTIONS
                and key not in SPLINE_APPROXIMATION_OPTIONS
            ):
                raise ValueError(f"Unknown inner option {key}.")

    def _get_quantitative_data_mask(
        self,
        petab_problem: 'petab.Problem',
        observable_ids: List[str],
        edatas: List['amici.ExpData'],
    ) -> List[np.ndarray]:
        # transform experimental data
        edatas = [
            amici.numpy.ExpDataView(edata)['observedData'] for edata in edatas
        ]

        measurement_df = petab_problem.measurement_df

        quantitative_data_mask = [
            np.zeros_like(edata, dtype=bool) for edata in edatas
        ]

        for observable_idx, observable_id in enumerate(observable_ids):
            observable_df = measurement_df[
                measurement_df[OBSERVABLE_ID] == observable_id
            ]
            # If the MEASUREMENT_TYPE column is filled with nans,
            # then fill that axis of the mask with True
            if observable_df[MEASUREMENT_TYPE].isna().all():
                for condition_mask in quantitative_data_mask:
                    condition_mask[:, observable_idx] = True

        # If there is no quantitative data, return None
        if not all(mask.any() for mask in quantitative_data_mask):
            return None

        return quantitative_data_mask

    def __call__(
        self,
        x_dct: Dict,
        sensi_orders: Tuple[int],
        mode: ModeType,
        amici_model: AmiciModel,
        amici_solver: AmiciSolver,
        edatas: List[amici.ExpData],
        n_threads: int,
        x_ids: Sequence[str],
        parameter_mapping: ParameterMapping,
        fim_for_hess: bool,
    ):
        """Perform the actual AMICI call.

        Called within the :func:`AmiciObjective.__call__` method.
        Calls all the inner calculators and combines the results.

        Parameters
        ----------
        x_dct:
            Parameters for which to compute function value and derivatives.
        sensi_orders:
            Tuple of requested sensitivity orders.
        mode:
            Call mode (function value or residual based).
        amici_model:
            The AMICI model.
        amici_solver:
            The AMICI solver.
        edatas:
            The experimental data.
        n_threads:
            Number of threads for AMICI call.
        x_ids:
            Ids of optimization parameters.
        parameter_mapping:
            Mapping of optimization to simulation parameters.
        fim_for_hess:
            Whether to use the FIM (if available) instead of the Hessian (if
            requested).
        """
        import amici.parameter_mapping

        if mode == MODE_RES:
            raise NotImplementedError(
                f"Mode {mode} is not implemented for the :class:`pypesto.objective.amici.InnerCalculatorCollector`."
            )

        if 2 in sensi_orders:
            raise ValueError(
                "Hessian and FIM are not implemented for the :class:`pypesto.objective.amici.InnerCalculatorCollector`."
            )

        # get dimension of outer problem
        dim = len(x_ids)

        # initialize return values
        nllh, snllh, s2nllh, chi2, res, sres = init_return_values(
            sensi_orders, mode, dim
        )
        inner_parameters_dictionary = {}

        # set order in solver
        sensi_order = 0
        if sensi_orders:
            sensi_order = max(sensi_orders)

        amici_solver.setSensitivityOrder(sensi_order)

        x_dct = copy.deepcopy(x_dct)

        # fill in parameters
        amici.parameter_mapping.fill_in_parameters(
            edatas=edatas,
            problem_parameters=x_dct,
            scaled_parameters=True,
            parameter_mapping=parameter_mapping,
            amici_model=amici_model,
        )

        # run amici simulation
        rdatas = amici.runAmiciSimulations(
            amici_model,
            amici_solver,
            edatas,
            num_threads=min(n_threads, len(edatas)),
        )

        # if any amici simulation failed, it's unlikely we can compute
        # meaningful inner parameters, so we better just fail early.
        if any(rdata.status != amici.AMICI_SUCCESS for rdata in rdatas):
            ret = {
                FVAL: nllh,
                GRAD: snllh,
                HESS: s2nllh,
                RES: res,
                SRES: sres,
                RDATAS: rdatas,
                X_INNER_OPT: inner_parameters_dictionary,
            }
            ret[FVAL] = np.inf
            # if the gradient was requested,
            # we need to provide some value for it
            if 1 in sensi_orders:
                ret[GRAD] = np.full(shape=len(x_ids), fill_value=np.nan)
            return filter_return_dict(ret)

        if (
            not self._known_least_squares_safe
            and mode == MODE_RES
            and 1 in sensi_orders
        ):
            if not amici_model.getAddSigmaResiduals() and any(
                (
                    (r[AMICI_SSIGMAY] is not None and np.any(r[AMICI_SSIGMAY]))
                    or (
                        r[AMICI_SSIGMAZ] is not None
                        and np.any(r[AMICI_SSIGMAZ])
                    )
                )
                for r in rdatas
            ):
                raise RuntimeError(
                    'Cannot use least squares solver with'
                    'parameter dependent sigma! Support can be '
                    'enabled via '
                    'amici_model.setAddSigmaResiduals().'
                )
            self._known_least_squares_safe = True  # don't check this again

        # call inner calculators
        for calculator in self.inner_calculators:
            inner_result = calculator(
                x_dct=x_dct,
                sensi_orders=sensi_orders,
                mode=mode,
                amici_model=amici_model,
                amici_solver=amici_solver,
                edatas=edatas,
                n_threads=n_threads,
                x_ids=x_ids,
                parameter_mapping=parameter_mapping,
                fim_for_hess=fim_for_hess,
                rdatas=rdatas,
            )
            nllh += inner_result[FVAL]
            if sensi_order > 0:
                snllh += inner_result[GRAD]
            inner_parameters_dictionary.update(inner_result[X_INNER_OPT])

        # add result for quantitative data
        if self.quantitative_data_mask is not None:
            quantitative_result = calculate_quantitative_result(
                rdatas=rdatas,
                sensi_orders=sensi_orders,
                edatas=edatas,
                mode=mode,
                quantitative_data_mask=self.quantitative_data_mask,
                dim=dim,
                parameter_mapping=parameter_mapping,
                par_opt_ids=x_ids,
                par_sim_ids=amici_model.getParameterIds(),
                par_edatas_indices=[edata.plist for edata in edatas],
            )
            nllh += quantitative_result[FVAL]
            if sensi_order > 0:
                snllh += quantitative_result[GRAD]

        ret = {
            FVAL: nllh,
            GRAD: snllh,
            HESS: s2nllh,
            RES: res,
            SRES: sres,
            RDATAS: rdatas,
            X_INNER_OPT: inner_parameters_dictionary,
        }

        return filter_return_dict(ret)


def calculate_quantitative_result(
    rdatas: List[amici.ReturnDataView],
    edatas: List[amici.ExpData],
    sensi_orders: Tuple[int],
    mode: ModeType,
    quantitative_data_mask: List[np.ndarray],
    dim: int,
    parameter_mapping: ParameterMapping,
    par_opt_ids: List[str],
    par_sim_ids: List[str],
    par_edatas_indices: List[List[int]],
):
    """Calculate the function values from rdatas and return as dict."""
    nllh, snllh, s2nllh, chi2, res, sres = init_return_values(
        sensi_orders, mode, dim
    )

    # transform experimental data
    edatas = [
        amici.numpy.ExpDataView(edata)['observedData'] for edata in edatas
    ]

    # calculate the function value
    for rdata, edata, mask in zip(rdatas, edatas, quantitative_data_mask):
        data_i = edata[mask]
        sim_i = rdata[AMICI_Y][mask]
        sigma_i = rdata[AMICI_SIGMAY][mask]

        nllh += 0.5 * np.nansum(
            np.log(2 * np.pi * sigma_i**2)
            + (data_i - sim_i) ** 2 / sigma_i**2
        )

    # calculate the gradient if requested
    if 1 in sensi_orders:
        parameter_map_sim_var = [
            cond_par_map.map_sim_var for cond_par_map in parameter_mapping
        ]
        # iterate over simulation conditions
        for (
            rdata,
            edata,
            mask,
            condition_map_sim_var,
            par_edata_indices,
        ) in zip(
            rdatas,
            edatas,
            quantitative_data_mask,
            parameter_map_sim_var,
            par_edatas_indices,
        ):
            data_i = edata[mask]
            sim_i = rdata[AMICI_Y][mask]
            sigma_i = rdata[AMICI_SIGMAY][mask]

            n_parameters = rdata[AMICI_SY].shape[1]

            # Get sensitivities of observables and sigmas
            sensitivities_i = np.asarray(
                [
                    rdata[AMICI_SY][:, parameter_index, :][mask]
                    for parameter_index in range(n_parameters)
                ]
            )
            ssigma_i = np.asarray(
                [
                    rdata[AMICI_SSIGMAY][:, parameter_index][mask]
                    for parameter_index in range(n_parameters)
                ]
            )
            # calculate the gradient for the condition
            gradient_for_condition = ssigma_i @ (
                (
                    np.full(len(data_i), 1)
                    - (data_i - sim_i) ** 2 / sigma_i**2
                )
                / sigma_i
            ) - sensitivities_i @ ((data_i - sim_i) / sigma_i**2)

            # add gradient to correct index of snllh
            for par_sim, par_opt in condition_map_sim_var.items():
                if not isinstance(par_opt, str):
                    continue
                par_opt_idx = par_opt_ids.index(par_opt)
                par_sim_idx = par_sim_ids.index(par_sim)
                par_edata_idx = (
                    par_edata_indices.index(par_sim_idx)
                    if par_sim_idx in par_edata_indices
                    else None
                )

                if par_edata_idx is not None:
                    snllh[par_opt_idx] += gradient_for_condition[par_edata_idx]

    ret = {
        FVAL: nllh,
        GRAD: snllh,
        HESS: s2nllh,
        RES: res,
        SRES: sres,
        RDATAS: rdatas,
    }
    return filter_return_dict(ret)



"""Engines with multi-threading parallelization."""
import copy
import logging
import os
from concurrent.futures import ThreadPoolExecutor
from typing import Any, List

from tqdm import tqdm

from .base import Engine
from .task import Task

logger = logging.getLogger(__name__)


def work(task):
    """Just execute task."""
    return task.execute()


class MultiThreadEngine(Engine):
    """
    Parallelize the task execution using multithreading.

    Parameters
    ----------
    n_threads:
        The maximum number of threads to use in parallel.
        Defaults to the number of CPUs available on the system according to
        `os.cpu_count()`.
        The effectively used number of threads will be the minimum of
        `n_threads` and the number of tasks submitted.
    """

    def __init__(self, n_threads: int = None):
        super().__init__()

        if n_threads is None:
            n_threads = os.cpu_count()
            logger.info(
                f"Engine will use up to {n_threads} threads (= CPU count)."
            )
        self.n_threads: int = n_threads

    def execute(
        self, tasks: List[Task], progress_bar: bool = True
    ) -> List[Any]:
        """Deepcopy tasks and distribute work over parallel threads.

        Parameters
        ----------
        tasks:
            List of tasks to execute.
        progress_bar:
            Whether to display a progress bar.
        """
        n_tasks = len(tasks)

        copied_tasks = [copy.deepcopy(task) for task in tasks]

        n_threads = min(self.n_threads, n_tasks)
        logger.debug(f"Parallelizing on {n_threads} threads.")

        with ThreadPoolExecutor(max_workers=n_threads) as pool:
            results = list(
                tqdm(
                    pool.map(work, copied_tasks),
                    total=len(copied_tasks),
                    disable=not progress_bar,
                ),
            )

        return results



"""DynestySampler class."""

# TODO set `nlive` to > a few * n_params**2
#      - https://github.com/joshspeagle/dynesty/issues/301#issuecomment-888461184
#      - seems easy for `NestedSampler(..., live_points=...)`, and for
#        `DynamicNestedSampler(..., nlive0=...)`.
# TODO set one of the initial live points to the best vector from optimization
#      - this should hopefully ensure a good estimate for `L_max`, used for
#        stopping criteria
#      - seems easy for `NestedSampler(..., live_points)`, but not for
#        `DynamicNestedSampler`.
# TODO allow selection of bounding methods
#      - which bounding method works well for high-dimension problems?
#        - possibly overlapping balls/cubes (reduced tuning parameters), or
#          unit cube (but only if unimodal?)
#          - ellipsoids seem to be recommended
#        - requires benchmarking...
# TODO consider "inner" sampling method
#      - Hamiltonian for large problems (D > 20) with gradients
#        - need to reshape gradients to unit cube
#        - need to provide gradients, hopefully without additional obj call
#        - or use `compute_jac` option, probably expensive
#      - multivariate slice for large problems (D > 20) without gradients
# TODO consider reflective boundaries

from __future__ import annotations

import logging
from typing import List, Union

import numpy as np

from ..problem import Problem
from ..result import McmcPtResult
from .sampler import Sampler, SamplerImportError

logger = logging.getLogger(__name__)


class DynestySampler(Sampler):
    """Use dynesty for sampling.

    NB: `get_samples` returns MCMC-like samples, by resampling original
    `dynesty` samples according to their importance weights. This is because
    the original samples contain many low-likelihood samples.
    To work with the original samples, modify the results object with
    `pypesto_result.sample_result = sampler.get_original_samples()`, where
    `sampler` is an instance of `pypesto.sample.DynestySampler`. The original
    dynesty results object is available at `sampler.results`.

    NB: the dynesty samplers can be customized significantly, by providing
    `sampler_args` and `run_args` to your `pypesto.sample.DynestySampler()`
    call. For example, code to parallelize dynesty is provided in pyPESTO's
    `sampler_study.ipynb` notebook.

    Wrapper around https://dynesty.readthedocs.io/en/stable/, see there for
    details.
    """

    def __init__(
        self,
        sampler_args: dict = None,
        run_args: dict = None,
        dynamic: bool = True,
    ):
        """
        Initialize sampler.

        Parameters
        ----------
        sampler_args:
            Further keyword arguments that are passed on to the `__init__`
            method of the dynesty sampler.
        run_args:
            Further keyword arguments that are passed on to the `run_nested`
            method of the dynesty sampler.
        dynamic:
            Whether to use dynamic or static nested sampling.
        """
        # check dependencies
        import dynesty

        setup_dynesty()

        super().__init__()

        self.dynamic = dynamic

        if sampler_args is None:
            sampler_args = {}
        self.sampler_args: dict = sampler_args

        if run_args is None:
            run_args = {}
        self.run_args: dict = run_args

        # set in initialize
        self.problem: Union[Problem, None] = None
        self.sampler: Union[
            dynesty.DynamicNestedSampler,
            dynesty.NestedSampler,
            None,
        ] = None

    def prior_transform(self, prior_sample: np.ndarray) -> np.ndarray:
        """Transform prior sample from unit cube to pyPESTO prior.

        TODO support priors that are not uniform.
             raise warning in `self.initialize` for now.

        Parameters
        ----------
        prior_sample:
            The prior sample, provided by dynesty.
        problem:
            The pyPESTO problem.

        Returns
        -------
        The transformed prior sample.
        """
        return (
            prior_sample * (self.problem.ub - self.problem.lb)
            + self.problem.lb
        )

    def loglikelihood(self, x):
        """Log-probability density function."""
        # check if parameter lies within bounds
        if any(x < self.problem.lb) or any(x > self.problem.ub):
            return -np.inf
        # invert sign
        # TODO this is possibly the posterior if priors are defined
        return -1.0 * self.problem.objective(x)

    def initialize(
        self,
        problem: Problem,
        x0: Union[np.ndarray, List[np.ndarray]],
    ) -> None:
        """Initialize the sampler."""
        import dynesty

        setup_dynesty()

        self.problem = problem

        sampler_class = dynesty.NestedSampler
        if self.dynamic:
            sampler_class = dynesty.DynamicNestedSampler

        # initialize sampler
        self.sampler = sampler_class(
            loglikelihood=self.loglikelihood,
            prior_transform=self.prior_transform,
            ndim=len(self.problem.x_free_indices),
            **self.sampler_args,
        )

        # TODO somehow use optimized vector as one of the initial live points

    def sample(self, n_samples: int, beta: float = None) -> None:
        """Return the most recent sample state."""
        if n_samples is not None:
            logger.warning(
                "`n_samples` was specified but this is incompatible with "
                "`dynesty` samplers, as they run until other stopping "
                "criteria are satisfied."
            )
        if beta is not None:
            logger.warning(
                "The temperature of a `dynesty` sampler was set, but this is "
                "irrelevant for `dynesty` samplers."
            )

        self.sampler.run_nested(**self.run_args)
        self.results = self.sampler.results

    def save_internal_sampler(self, filename: str) -> None:
        """Save the state of the internal dynesty sampler.

        This makes it easier to analyze the original dynesty samples, after
        sampling, with `restore_internal`.

        Parameters
        ----------
        filename:
            The internal sampler will be saved here.
        """
        import dynesty

        setup_dynesty()

        dynesty.utils.save_sampler(
            sampler=self.sampler,
            fname=filename,
        )

    def restore_internal_sampler(self, filename: str) -> None:
        """Restore the state of the internal dynesty sampler.

        Parameters
        ----------
        filename:
            The internal sampler will be saved here.
        """
        import dynesty

        setup_dynesty()

        self.sampler = dynesty.utils.restore_sampler(fname=filename)

    def get_original_samples(self) -> McmcPtResult:
        """Get the samples into the fitting pypesto format.

        Returns
        -------
        The pyPESTO sample result.
        """
        return get_original_dynesty_samples(sampler=self.sampler)

    def get_samples(self) -> McmcPtResult:
        """Get MCMC-like samples into the fitting pypesto format.

        Returns
        -------
        The pyPESTO sample result.
        """
        return get_mcmc_like_dynesty_samples(sampler=self.sampler)


def get_original_dynesty_samples(sampler) -> McmcPtResult:
    """Get original dynesty samples.

    Parameters
    ----------
    sampler:
        The (internal!) dynesty sampler. See
        `pypesto.sample.DynestySampler.__init__`, specifically the
        `save_internal` argument, for more details.

    Returns
    -------
    The sample result.
    """
    trace_x = np.array([sampler.results.samples])
    trace_neglogpost = -np.array([sampler.results.logl])

    # the sampler uses custom adaptive priors
    trace_neglogprior = np.full(trace_neglogpost.shape, np.nan)
    # the sampler uses temperature 1
    betas = np.array([1.0])

    result = McmcPtResult(
        trace_x=trace_x,
        trace_neglogpost=trace_neglogpost,
        trace_neglogprior=trace_neglogprior,
        betas=betas,
    )

    return result


def get_mcmc_like_dynesty_samples(sampler) -> McmcPtResult:
    """Get MCMC-like samples.

    Parameters
    ----------
    sampler:
        The (internal!) dynesty sampler. See
        `pypesto.sample.DynestySampler.__init__`, specifically the
        `save_internal` argument, for more details.

    Returns
    -------
    The sample result.
    """
    import dynesty

    setup_dynesty()

    if len(sampler.results.importance_weights().shape) != 1:
        raise ValueError(
            "Unknown error. The dynesty importance weights are not a 1D array."
        )
    # resample according to importance weights
    indices = dynesty.utils.resample_equal(
        np.arange(sampler.results.importance_weights().shape[0]),
        sampler.results.importance_weights(),
    )

    trace_x = np.array([sampler.results.samples[indices]])
    trace_neglogpost = -np.array([sampler.results.logl[indices]])

    trace_neglogprior = np.array([np.full((len(indices),), np.nan)])
    betas = np.array([1.0])

    result = McmcPtResult(
        trace_x=trace_x,
        trace_neglogpost=trace_neglogpost,
        trace_neglogprior=trace_neglogprior,
        betas=betas,
    )
    return result


def setup_dynesty() -> None:
    """Import dynesty."""
    try:
        import dill  # noqa: S403
        import dynesty.utils

        dynesty.utils.pickle_module = dill
    except ImportError:
        raise SamplerImportError("dynesty")



"""Definition of an optimal scaling inner parameter class."""
import logging
from typing import Literal

from ...C import (
    INTERVAL_CENSORED,
    LEFT_CENSORED,
    RIGHT_CENSORED,
    InnerParameterType,
)
from ..parameter import InnerParameter

logger = logging.getLogger(__name__)


class OptimalScalingParameter(InnerParameter):
    """Inner parameter of the optimal scaling hierarchical optimization problem.

    Attributes
    ----------
    dummy_value:
        Value to be used when the optimal parameter is not yet known
        (in particular to simulate unscaled observables).
    inner_parameter_id:
        The inner parameter ID.
    inner_parameter_type:
        The inner parameter type.
    ixs:
        A mask (boolean matrix) that indicates the measurements that this
        parameter is used in.
    lb:
        The lower bound, for optimization.
    scale:
        Scale on which to estimate this parameter.
    ub:
        The upper bound, for optimization.
    observable_id:
        The id of the observable whose measurements are in the category
        with this inner parameter.
    category:
        Category index.
    group:
        Group index.
    value:
        Current value of the inner parameter.
    estimate:
        Whether to estimate inner parameter in inner subproblem.
    censoring_type:
        The censoring type of the measurements in the category with this
        inner parameter. In case of ordinal measurements, this is None.
    """

    def __init__(
        self,
        *args,
        observable_id: str = None,
        group: int = None,
        category: int = None,
        estimate: bool = False,
        censoring_type: Literal[
            None, LEFT_CENSORED, RIGHT_CENSORED, INTERVAL_CENSORED
        ] = None,
        **kwargs,
    ):
        """Construct.

        Parameters
        ----------
        See class attributes.
        """
        super().__init__(*args, **kwargs)
        if self.inner_parameter_type != InnerParameterType.OPTIMAL_SCALING:
            raise ValueError(
                f"For the OptimalScalingParameter class, the parameter type has to be {InnerParameterType.OPTIMAL_SCALING}."
            )

        if group is None:
            raise ValueError("No Parameter group provided.")
        if category is None:
            raise ValueError("No Category provided.")

        self.observable_id = observable_id
        self.category = category
        self.group = group
        self.estimate = estimate
        self.value = self.dummy_value
        self.censoring_type = censoring_type



"""Definition of an optimal scaling solver class."""
import warnings
from typing import Dict, List, Tuple

import numpy as np

from ...C import (
    C_MATRIX,
    CENSORED,
    INTERVAL_CONSTRAINTS,
    LB_INDICES,
    MAX,
    MAXMIN,
    MEASUREMENT_TYPE,
    METHOD,
    MIN_GAP,
    NUM_CATEGORIES,
    NUM_DATAPOINTS,
    NUM_INNER_PARAMS,
    ORDINAL,
    QUANTITATIVE_DATA,
    QUANTITATIVE_IXS,
    REDUCED,
    REPARAMETERIZED,
    SCIPY_FUN,
    SCIPY_SUCCESS,
    SCIPY_X,
    STANDARD,
    SURROGATE_DATA,
    UB_INDICES,
    W_DOT_MATRIX,
    W_MATRIX,
    InnerParameterType,
)
from ..solver import InnerSolver
from .parameter import OptimalScalingParameter
from .problem import OptimalScalingProblem

try:
    from amici.parameter_mapping import ParameterMapping
except ImportError:
    pass


class OptimalScalingInnerSolver(InnerSolver):
    """Solve the inner subproblem of the optimal scaling approach for ordinal data.

    Options
    -------
    method:
        The method to use for the inner optimization problem.
        Can be 'standard' or 'reduced'. The latter is more efficient.
    reparameterized:
        Whether to use reparameterized optimization.
    intervalConstraints:
        The type of interval constraints to use.
        Can be 'max' or 'maxmin'.
    minGap:
        The minimum gap between two consecutive categories.
    """

    def __init__(self, options: Dict = None):
        """Construct."""
        self.options = {
            **self.get_default_options(),
            **(options or {}),
        }
        self.validate_options()

        self.x_guesses = None

    def validate_options(self):
        """Validate the current options dictionary."""
        if self.options[METHOD] not in [STANDARD, REDUCED]:
            raise ValueError(
                f"Inner solver method cannot be {self.options[METHOD]}. Please enter either {STANDARD} or {REDUCED}"
            )
        elif type(self.options[REPARAMETERIZED]) is not bool:
            raise ValueError(
                f"Inner solver option 'reparameterized' has to be boolean, not {type(self.options[REPARAMETERIZED])}."
            )
        elif self.options[INTERVAL_CONSTRAINTS] not in [MAX, MAXMIN]:
            raise ValueError(
                f"Inner solver method cannot be {self.options[INTERVAL_CONSTRAINTS]}. Please enter either {MAX} or {MAXMIN}"
            )
        elif type(self.options[MIN_GAP]) is not float:
            raise ValueError(
                f"Inner solver option 'reparameterized' has to be a float, not {type(self.options[MIN_GAP])}."
            )
        elif (
            self.options[METHOD] == STANDARD and self.options[REPARAMETERIZED]
        ):
            raise NotImplementedError(
                'Combining standard approach with '
                'reparameterization not implemented.'
            )
        elif self.options[METHOD] == STANDARD:
            warnings.warn(
                'Standard approach is not recommended, as it is less efficient.'
                'Please consider using the reduced approach instead.'
            )
        # Check for any other options
        for key in self.options:
            if key not in self.get_default_options():
                raise ValueError(
                    f"Unknown OptimalScalingInnerSolver option {key}."
                )

    def solve(
        self,
        problem: OptimalScalingProblem,
        sim: List[np.ndarray],
        sigma: List[np.ndarray],
    ) -> list:
        """Get results for every group (inner optimization problem).

        Parameters
        ----------
        problem:
            Optimal scaling inner problem.
        sim:
            Model simulations.
        sigma:
            Standard deviation of the noise.

        Returns
        -------
        List of optimization results of the inner subproblem.
        """
        optimal_surrogates = []
        for group in problem.get_groups_for_xs(
            InnerParameterType.OPTIMAL_SCALING
        ):
            category_upper_bounds = problem.get_cat_ub_parameters_for_group(
                group
            )
            category_lower_bounds = problem.get_cat_lb_parameters_for_group(
                group
            )
            if problem.groups[group][MEASUREMENT_TYPE] == ORDINAL:
                surrogate_opt_results = optimize_surrogate_data_per_group(
                    category_upper_bounds=category_upper_bounds,
                    category_lower_bounds=category_lower_bounds,
                    sim=sim,
                    options=self.options,
                )
                save_inner_parameters_to_inner_problem(
                    category_upper_bounds=category_upper_bounds,
                    group=group,
                    problem=problem,
                    x_inner_opt=surrogate_opt_results,
                    sim=sim,
                    options=self.options,
                )
            elif problem.groups[group][MEASUREMENT_TYPE] == CENSORED:
                quantitative_data = problem.groups[group][QUANTITATIVE_DATA]
                quantitative_ixs = problem.groups[group][QUANTITATIVE_IXS]
                surrogate_opt_results = calculate_censored_obj(
                    category_upper_bounds=category_upper_bounds,
                    category_lower_bounds=category_lower_bounds,
                    sim=sim,
                    sigma=sigma,
                    quantitative_data=quantitative_data,
                    quantitative_ixs=quantitative_ixs,
                )

            optimal_surrogates.append(surrogate_opt_results)
        return optimal_surrogates

    @staticmethod
    def calculate_obj_function(x_inner_opt: list):
        """Calculate the inner objective function value.

        Calculates the inner objective function value from a list of inner
        optimization results returned from `OptimalScalingInnerSolver.solve`.

        Parameters
        ----------
        x_inner_opt:
            List of optimization results of the inner subproblem.

        Returns
        -------
        Inner objective function value.
        """
        if False in [
            x_inner_opt[idx][SCIPY_SUCCESS] for idx in range(len(x_inner_opt))
        ]:
            obj = np.inf
            warnings.warn("Inner optimization failed.")
        else:
            obj = np.sum(
                [
                    x_inner_opt[idx][SCIPY_FUN]
                    for idx in range(len(x_inner_opt))
                ]
            )
        return obj

    def calculate_gradients(
        self,
        problem: OptimalScalingProblem,
        x_inner_opt: List[Dict],
        sim: List[np.ndarray],
        sy: List[np.ndarray],
        sigma: List[np.ndarray],
        ssigma: List[np.ndarray],
        parameter_mapping: ParameterMapping,
        par_opt_ids: List,
        par_sim_ids: List,
        par_edatas_indices: List,
        snllh: Dict,
    ):
        """Calculate gradients of the inner objective function.

        Calculates gradients of the objective function with respect to outer
        (dynamical) parameters.

        Parameters
        ----------
        problem:
            Optimal scaling inner problem.
        x_inner_opt:
            List of optimization results of the inner subproblem.
        sim:
            Model simulations.
        sy:
            Model sensitivities.
        sigma:
            Model noise parameters.
        ssigma:
            Model sensitivity of noise parameters.
        parameter_mapping:
            Mapping of optimization to simulation parameters.
        par_opt_ids:
            Ids of outer otimization parameters.
        par_sim_ids:
            Ids of outer simulation parameters, includes fixed parameters.
        par_edata_indices:
            Indices of parameters from `amici_model.getParameterIds()` that are needed for
            sensitivity calculation. Comes from `edata.plist` for each condition.
        snllh:
            Empty dictionary with optimization parameters as keys.

        Returns
        -------
        Filled in snllh dictionary with objective function gradients.
        """
        already_calculated = set()

        for condition_map_sim_var in [
            cond_par_map.map_sim_var for cond_par_map in parameter_mapping
        ]:
            # Iterate over outer optimization parameters.
            for par_sim, par_opt in condition_map_sim_var.items():
                if (
                    not isinstance(par_opt, str)
                    or par_opt in already_calculated
                ):
                    continue
                else:
                    already_calculated.add(par_opt)

                par_opt_idx = par_opt_ids.index(par_opt)
                par_sim_idx = par_sim_ids.index(par_sim)
                par_edata_idx = [
                    par_edata_indices.index(par_sim_idx)
                    if par_sim_idx in par_edata_indices
                    else None
                    for par_edata_indices in par_edatas_indices
                ]

                grad = 0.0

                # Iterate over inner parameter groups.
                for idx, group in enumerate(
                    problem.get_groups_for_xs(
                        InnerParameterType.OPTIMAL_SCALING
                    )
                ):
                    if problem.groups[group][MEASUREMENT_TYPE] == CENSORED:
                        category_upper_bounds = (
                            problem.get_cat_ub_parameters_for_group(group)
                        )
                        category_lower_bounds = (
                            problem.get_cat_lb_parameters_for_group(group)
                        )
                        quantitative_data = problem.groups[group][
                            QUANTITATIVE_DATA
                        ]
                        quantitative_ixs = problem.groups[group][
                            QUANTITATIVE_IXS
                        ]

                        grad += calculate_censored_grad(
                            category_upper_bounds=category_upper_bounds,
                            category_lower_bounds=category_lower_bounds,
                            sim=sim,
                            sy=sy,
                            sigma=sigma,
                            ssigma=ssigma,
                            par_edata_idx=par_edata_idx,
                            quantitative_data=quantitative_data,
                            quantitative_ixs=quantitative_ixs,
                        )
                    elif problem.groups[group][MEASUREMENT_TYPE] == ORDINAL:
                        xs = problem.get_cat_ub_parameters_for_group(group)
                        xi = get_xi(
                            group, problem, x_inner_opt[idx], sim, self.options
                        )
                        sim_all = get_sim_all(xs, sim)
                        sy_all = get_sy_all(xs, sy, par_edata_idx)

                        problem.groups[group][W_MATRIX] = problem.get_w(
                            group, sim_all
                        )
                        problem.groups[group][W_DOT_MATRIX] = problem.get_wdot(
                            group, sim_all, sy_all
                        )

                        residual = np.block(
                            [
                                xi[: problem.groups[group][NUM_DATAPOINTS]]
                                - sim_all,
                                np.zeros(
                                    problem.groups[group][NUM_INNER_PARAMS]
                                    - problem.groups[group][NUM_DATAPOINTS]
                                ),
                            ]
                        )
                        dy_dtheta = get_dy_dtheta(group, problem, sy_all)

                        df_dtheta = residual.dot(
                            residual.dot(problem.groups[group][W_DOT_MATRIX])
                            - 2
                            * problem.groups[group][W_MATRIX].dot(dy_dtheta)
                        )
                        df_dxi = 2 * problem.groups[group][W_MATRIX].dot(
                            residual
                        )

                        if df_dxi.any():
                            dd_dtheta = problem.get_dd_dtheta(
                                group, xs, sim_all, sy_all
                            )
                            d = problem.get_d(
                                group, xs, sim_all, self.options[MIN_GAP]
                            )

                            mu = get_mu(group, problem, residual)

                            dxi_dtheta = calculate_dxi_dtheta(
                                group,
                                problem,
                                xi,
                                mu,
                                dy_dtheta,
                                residual,
                                d,
                                dd_dtheta,
                            )

                            grad += dxi_dtheta.dot(df_dxi) + df_dtheta
                        else:
                            grad += df_dtheta

                snllh[par_opt_idx] = grad
        return snllh

    @staticmethod
    def get_default_options() -> Dict:
        """Return default options for solving the inner problem."""
        options = {
            METHOD: REDUCED,
            REPARAMETERIZED: True,
            INTERVAL_CONSTRAINTS: MAX,
            MIN_GAP: 1e-16,
        }
        return options


def calculate_dxi_dtheta(
    group: int,
    problem: OptimalScalingProblem,
    xi: np.ndarray,
    mu: np.ndarray,
    dy_dtheta: np.ndarray,
    residual: np.ndarray,
    d: np.ndarray,
    dd_dtheta: np.ndarray,
):
    """Calculate derivatives of inner parameters with respect to outer parameter.

    Parameters
    ----------
    group:
        Inner parameter group.
    problem:
        Optimal scaling inner problem.
    xi:
        Inner parameters: category bounds and surrogate data.
    mu:
        Lagrange multipliers of the inner optimization problem.
    dy_dtheta:
        Model sensitivities for group.
    residual:
        Residual for group.
    d:
        Vector of interval gap and range.
    dd_theta:
        Derivative of vector of interval gap and range.
    """
    from scipy.sparse import csc_matrix, linalg

    A = np.block(
        [
            [
                2 * problem.groups[group][W_MATRIX],
                problem.groups[group][C_MATRIX].transpose(),
            ],
            [
                (mu * problem.groups[group][C_MATRIX].transpose()).transpose(),
                np.diag(problem.groups[group][C_MATRIX].dot(xi) + d),
            ],
        ]
    )
    A_sp = csc_matrix(A)

    b = np.block(
        [
            2 * dy_dtheta.dot(problem.groups[group][W_MATRIX])
            - 2 * problem.groups[group][W_DOT_MATRIX].dot(residual),
            -mu * dd_dtheta,
        ]
    )

    dxi_dtheta = linalg.spsolve(A_sp, b)
    return dxi_dtheta[: problem.groups[group][NUM_INNER_PARAMS]]


def get_dy_dtheta(
    group: int, problem: OptimalScalingProblem, sy_all: np.ndarray
):
    """Restructure sensitivities into a numpy matrix of right dimension."""
    return np.block(
        [sy_all, np.zeros(2 * problem.groups[group][NUM_CATEGORIES])]
    )


def get_mu(group: int, problem: OptimalScalingProblem, residual: np.ndarray):
    """Calculate Lagrange multipliers of the inner optimization problem.

    Parameters
    ----------
    group:
        Inner parameter group.
    problem:
        Optimal scaling inner problem.
    residual:
        Residual for group.
    """
    from scipy import linalg

    mu = linalg.lstsq(
        problem.groups[group][C_MATRIX].transpose(),
        -2 * residual.dot(problem.groups[group][W_MATRIX]),
        lapack_driver='gelsy',
    )
    return mu[0]


def get_xi(
    group: int,
    problem: OptimalScalingProblem,
    x_inner_opt: Dict,
    sim: List[np.ndarray],
    options: Dict,
):
    """Extract and calculate category bounds and surrogate data.

    Parameters
    ----------
    group:
        Inner parameter group.
    problem:
        Optimal scaling inner problem.
    x_inner_opt:
        Optimization results of the inner optimization subproblem.
    sim:
        Model simulation.
    options:
        Optimal scaling inner solver options.
    """
    xs = problem.get_cat_ub_parameters_for_group(group)
    interval_range, interval_gap = compute_interval_constraints(
        xs, sim, options
    )

    xi = np.zeros(problem.groups[group][NUM_INNER_PARAMS])
    surrogate_all, x_lower, x_upper = get_surrogate_all(
        xs, x_inner_opt[SCIPY_X], sim, interval_range, interval_gap, options
    )
    xi[: problem.groups[group][NUM_DATAPOINTS]] = surrogate_all.flatten()
    xi[problem.groups[group][LB_INDICES]] = x_lower
    xi[problem.groups[group][UB_INDICES]] = x_upper
    return xi


def optimize_surrogate_data_per_group(
    category_upper_bounds: List[OptimalScalingParameter],
    category_lower_bounds: List[OptimalScalingParameter],
    sim: List[np.ndarray],
    options: Dict,
):
    """Run optimization for inner subproblem.

    Parameters
    ----------
    category_upper_bounds:
        Upper bound parameters of categories for this group.
    category_lower_bounds:
        Lower bound parameters of categories for this group.
    sim:
        Model simulations.
    options:
        Optimal scaling inner solver options.
    """
    from scipy.optimize import minimize

    interval_range, interval_gap = compute_interval_constraints(
        category_upper_bounds, sim, options
    )
    w = get_weight_for_surrogate(category_upper_bounds, sim)

    def obj_surr(x):
        return obj_surrogate_data(
            category_upper_bounds,
            x,
            sim,
            interval_gap,
            interval_range,
            w,
            options,
        )

    def grad_surr(x):
        return grad_surrogate_data(
            category_upper_bounds,
            x,
            sim,
            interval_gap,
            interval_range,
            w,
            options,
        )

    inner_options = get_inner_optimization_options(
        category_upper_bounds,
        category_lower_bounds,
        sim,
        interval_range,
        interval_gap,
        options,
    )
    try:
        results = minimize(obj_surr, jac=grad_surr, **inner_options)
    except ValueError:
        warnings.warn(
            "x0 violate bound constraints. Retrying with array of zeros."
        )
        inner_options['x0'] = np.zeros(len(inner_options['x0']))
        results = minimize(obj_surr, jac=grad_surr, **inner_options)

    return results


def get_inner_optimization_options(
    category_upper_bounds: List[OptimalScalingParameter],
    category_lower_bounds: List[OptimalScalingParameter],
    sim: List[np.ndarray],
    interval_range: float,
    interval_gap: float,
    options: Dict,
) -> Dict:
    """Return default options for scipy optimizer.

    Returns inner subproblem optimization options including startpoint
    and optimization bounds or constraints, dependent on solver method.

    Parameters
    ----------
    category_upper_bounds:
        Upper bound parameters of categories for this group.
    category_lower_bounds:
        Lower bound parameters of categories for this group.
    sim:
        Model simulations.
    interval_range:
        Minimal constrained category interval range.
    interval_gap:
        Minimal constrained gap between categories.
    options:
        Optimal scaling inner solver options.
    """
    from scipy.optimize import Bounds

    min_all, max_all = get_min_max(category_upper_bounds, sim)
    if options[METHOD] == REDUCED:
        last_opt_values = np.asarray([x.value for x in category_upper_bounds])
    elif options[METHOD] == STANDARD:
        last_opt_values = np.ravel(
            [
                np.asarray([x.value for x in category_lower_bounds]),
                np.asarray([x.value for x in category_upper_bounds]),
            ],
            'F',
        )

    if options[METHOD] == REDUCED:
        parameter_length = len(category_upper_bounds)
        if len(np.nonzero(last_opt_values)) > 0:
            x0 = last_opt_values
        else:
            x0 = np.linspace(
                np.max([min_all, interval_range]),
                max_all + (interval_range + interval_gap) * parameter_length,
                parameter_length,
            )
    elif options[METHOD] == STANDARD:
        parameter_length = 2 * len(category_upper_bounds)
        if len(np.nonzero(last_opt_values)) > 0:
            x0 = last_opt_values
        else:
            x0 = np.linspace(0, max_all + interval_range, parameter_length)
    else:
        raise NotImplementedError(
            f"Unknown optimal scaling method {options[METHOD]}. "
            f"Please use {STANDARD} or {REDUCED}."
        )

    if options[REPARAMETERIZED]:
        x0 = reparameterize_inner_parameters(
            x0, category_upper_bounds, interval_gap, interval_range
        )
        # If taking last_opt_values, it's possible that x0 is negative.
        # due to different simulations. Clip so bounds are satisfied.
        x0 = np.clip(x0, 0, None)

        bounds = Bounds(
            [0.0] * parameter_length,
            [max_all + (interval_range + interval_gap) * parameter_length]
            * parameter_length,
        )
        inner_options = {
            'x0': x0,
            'method': 'L-BFGS-B',
            'options': {'maxiter': 2000, 'ftol': 1e-10},
            'bounds': bounds,
        }
    else:
        constraints = get_constraints_for_optimization(
            category_upper_bounds, sim, options
        )

        inner_options = {
            'x0': x0,
            'method': 'SLSQP',
            'options': {'maxiter': 2000, 'ftol': 1e-10, 'disp': None},
            'constraints': constraints,
        }
    return inner_options


def get_min_max(
    inner_parameters: List[OptimalScalingParameter], sim: List[np.ndarray]
) -> Tuple[float, float]:
    """Return minimal and maximal simulation value."""
    sim_all = get_sim_all(inner_parameters, sim)

    min_all = np.min(sim_all)
    max_all = np.max(sim_all)

    return min_all, max_all


def get_sy_all(
    inner_parameters: List[OptimalScalingParameter],
    sy: List[np.ndarray],
    par_edata_idx: List,
):
    """Return model sensitivities for inner parameters and outer parameter index."""
    sy_all = []
    for inner_parameter in inner_parameters:
        for sy_i, mask_i, edata_idx in zip(
            sy, inner_parameter.ixs, par_edata_idx
        ):
            if edata_idx is not None:
                sim_sy = sy_i[:, edata_idx, :][mask_i]
            else:
                sim_sy = np.full(sy_i[:, 0, :][mask_i].shape, 0)
            for sim_sy_i in sim_sy:
                sy_all.append(sim_sy_i)
    return np.array(sy_all)


def get_sim_all(inner_parameters, sim: List[np.ndarray]) -> list:
    """Return model simulations for inner parameters."""
    sim_all = []
    for inner_parameter in inner_parameters:
        for sim_i, mask_i in zip(sim, inner_parameter.ixs):
            sim_x = sim_i[mask_i]
            for sim_x_i in sim_x:
                sim_all.append(sim_x_i)
    return sim_all


def get_surrogate_all(
    inner_parameters: List[OptimalScalingParameter],
    optimal_scaling_bounds: List,
    sim: List[np.ndarray],
    interval_range: float,
    interval_gap: float,
    options: Dict,
):
    """Return surrogate data, lower and upper category bounds."""
    if options[REPARAMETERIZED]:
        optimal_scaling_bounds = undo_inner_parameter_reparameterization(
            optimal_scaling_bounds,
            inner_parameters,
            interval_gap,
            interval_range,
        )
    surrogate_all = []
    x_lower_all = []
    x_upper_all = []
    for inner_parameter in inner_parameters:
        upper_bound, lower_bound = get_bounds_for_category(
            inner_parameter, optimal_scaling_bounds, interval_gap, options
        )
        for sim_i, mask_i in zip(sim, inner_parameter.ixs):
            y_sim = sim_i[mask_i]
            for y_sim_i in y_sim:
                if lower_bound > y_sim_i:
                    y_surrogate = lower_bound
                elif y_sim_i > upper_bound:
                    y_surrogate = upper_bound
                elif lower_bound <= y_sim_i <= upper_bound:
                    y_surrogate = y_sim_i
                else:
                    continue
                surrogate_all.append(y_surrogate)
        x_lower_all.append(lower_bound)
        x_upper_all.append(upper_bound)
    return (
        np.array(surrogate_all),
        np.array(x_lower_all),
        np.array(x_upper_all),
    )


def get_weight_for_surrogate(
    inner_parameters: List[OptimalScalingParameter], sim: List[np.ndarray]
) -> float:
    """Calculate weights for objective function."""
    sim_x_all = get_sim_all(inner_parameters, sim)
    eps = 1e-8

    w = np.sum(np.abs(sim_x_all)) + eps

    return w


def compute_interval_constraints(
    inner_parameters: List[OptimalScalingParameter],
    sim: List[np.ndarray],
    options: Dict,
) -> Tuple[float, float]:
    """Compute minimal interval range and gap."""
    # compute constraints on interval size and interval gap size
    # similar to Pargett et al. (2014)
    if MIN_GAP not in options:
        eps = 1e-16
    else:
        eps = options[MIN_GAP]

    min_simulation, max_simulation = get_min_max(inner_parameters, sim)

    if options[INTERVAL_CONSTRAINTS] == MAXMIN:
        interval_range = (max_simulation - min_simulation) / (
            2 * len(inner_parameters) + 1
        )
        interval_gap = (max_simulation - min_simulation) / (
            4 * (len(inner_parameters) - 1) + 1
        )
    elif options[INTERVAL_CONSTRAINTS] == MAX:
        interval_range = max_simulation / (2 * len(inner_parameters) + 1)
        interval_gap = max_simulation / (4 * (len(inner_parameters) - 1) + 1)
    else:
        raise NotImplementedError(
            f"intervalConstraints = "
            f"{options[INTERVAL_CONSTRAINTS]} not implemented. "
            f"Please use {MAX} or {MAXMIN}."
        )
    return interval_range, interval_gap + eps


def reparameterize_inner_parameters(
    original_inner_parameter_values: np.ndarray,
    inner_parameters: List[OptimalScalingParameter],
    interval_gap: float,
    interval_range: float,
) -> np.ndarray:
    """Transform original inner parameters to reparameterized inner parameters."""
    reparameterized_inner_parameter_values = np.full(
        shape=(np.shape(original_inner_parameter_values)), fill_value=np.nan
    )
    for inner_parameter in inner_parameters:
        inner_parameter_category = int(inner_parameter.category)
        if inner_parameter_category == 1:
            reparameterized_inner_parameter_values[
                inner_parameter_category - 1
            ] = (
                original_inner_parameter_values[inner_parameter_category - 1]
                - interval_range
            )
        else:
            reparameterized_inner_parameter_values[
                inner_parameter_category - 1
            ] = (
                original_inner_parameter_values[inner_parameter_category - 1]
                - original_inner_parameter_values[inner_parameter_category - 2]
                - interval_gap
                - interval_range
            )

    return reparameterized_inner_parameter_values


def undo_inner_parameter_reparameterization(
    reparameterized_inner_parameter_values: np.ndarray,
    inner_parameters: List[OptimalScalingParameter],
    interval_gap: float,
    interval_range: float,
) -> np.ndarray:
    """Transform reparameterized inner parameters to original inner parameters."""
    original_inner_parameter_values = np.full(
        shape=(np.shape(reparameterized_inner_parameter_values)),
        fill_value=np.nan,
    )
    for inner_parameter in inner_parameters:
        inner_parameter_category = int(inner_parameter.category)
        if inner_parameter_category == 1:
            original_inner_parameter_values[inner_parameter_category - 1] = (
                interval_range
                + reparameterized_inner_parameter_values[
                    inner_parameter_category - 1
                ]
            )
        else:
            original_inner_parameter_values[inner_parameter_category - 1] = (
                reparameterized_inner_parameter_values[
                    inner_parameter_category - 1
                ]
                + interval_gap
                + interval_range
                + original_inner_parameter_values[inner_parameter_category - 2]
            )
    return original_inner_parameter_values


def obj_surrogate_data(
    xs: List[OptimalScalingParameter],
    optimal_scaling_bounds: np.ndarray,
    sim: List[np.ndarray],
    interval_gap: float,
    interval_range: float,
    w: float,
    options: Dict,
) -> float:
    """Compute optimal scaling objective function."""
    obj = 0.0
    if options[REPARAMETERIZED]:
        optimal_scaling_bounds = undo_inner_parameter_reparameterization(
            optimal_scaling_bounds, xs, interval_gap, interval_range
        )

    for x in xs:
        x_upper, x_lower = get_bounds_for_category(
            x, optimal_scaling_bounds, interval_gap, options
        )
        for sim_i, mask_i in zip(sim, x.ixs):
            y_sim = sim_i[mask_i]
            for y_sim_i in y_sim:
                if x_lower > y_sim_i:
                    y_surrogate = x_lower
                elif y_sim_i > x_upper:
                    y_surrogate = x_upper
                elif x_lower <= y_sim_i <= x_upper:
                    y_surrogate = y_sim_i
                else:
                    continue
                obj += (y_surrogate - y_sim_i) ** 2
    obj = np.divide(obj, w)
    return obj


def grad_surrogate_data(
    xs: List[OptimalScalingParameter],
    optimal_scaling_bounds: np.ndarray,
    sim: List[np.ndarray],
    interval_gap: float,
    interval_range: float,
    w: float,
    options: Dict,
) -> float:
    """Compute optimal scaling objective function."""
    grad = np.zeros(len(optimal_scaling_bounds))
    if options[REPARAMETERIZED]:
        optimal_scaling_bounds = undo_inner_parameter_reparameterization(
            optimal_scaling_bounds, xs, interval_gap, interval_range
        )

    if options[METHOD] == STANDARD:
        for x in xs:
            x_category = int(x.category)
            x_lower = optimal_scaling_bounds[2 * x_category - 2]
            x_upper = optimal_scaling_bounds[2 * x_category - 1]

            for sim_i, mask_i in zip(sim, x.ixs):
                y_sim = sim_i[mask_i]
                for y_sim_i in y_sim:
                    if x_lower > y_sim_i:
                        y_surrogate = x_lower
                        grad[2 * x_category - 2] += 2 * (y_surrogate - y_sim_i)
                    elif y_sim_i > x_upper:
                        y_surrogate = x_upper
                        grad[2 * x_category - 1] += 2 * (y_surrogate - y_sim_i)
                    else:
                        continue
    elif options[METHOD] == REDUCED:
        for x in xs:
            x_category = int(x.category)
            x_upper = optimal_scaling_bounds[x_category - 1]
            if x_category == 1:
                x_lower = 0.0
            else:
                x_lower = optimal_scaling_bounds[x_category - 2] + interval_gap

            for sim_i, mask_i in zip(sim, x.ixs):
                y_sim = sim_i[mask_i]
                for y_sim_i in y_sim:
                    if x_lower > y_sim_i and x_category != 1:
                        y_surrogate = x_lower
                        grad[x_category - 2] += 2 * (y_surrogate - y_sim_i)
                    elif y_sim_i > x_upper:
                        y_surrogate = x_upper
                        grad[x_category - 1] += 2 * (y_surrogate - y_sim_i)
                    else:
                        continue

    if options[REPARAMETERIZED]:
        grad = reparameterize_gradient(
            grad,
            xs,
        )
    grad = np.divide(grad, w)
    return grad


def reparameterize_gradient(
    grad: np.ndarray,
    xs: List[OptimalScalingParameter],
) -> np.ndarray:
    """Transform gradient to reparameterized gradient."""
    reparameterized_grad = np.full(
        shape=(np.shape(grad)),
        fill_value=np.nan,
    )
    for inner_parameter in xs:
        inner_parameter_category = int(inner_parameter.category)
        reparameterized_grad[inner_parameter_category - 1] = np.sum(
            grad[inner_parameter_category - 1 :]
        )
    return reparameterized_grad


def get_bounds_for_category(
    x: OptimalScalingParameter,
    optimal_scaling_bounds: np.ndarray,
    interval_gap: float,
    options: Dict,
) -> Tuple[float, float]:
    """Return upper and lower bound for a specific category x."""
    x_category = int(x.category)

    if options[METHOD] == REDUCED:
        x_upper = optimal_scaling_bounds[x_category - 1]
        if x_category == 1:
            x_lower = 0
        elif x_category > 1:
            x_lower = optimal_scaling_bounds[x_category - 2] + interval_gap
        else:
            raise ValueError('Category value needs to be larger than 0.')
    elif options[METHOD] == STANDARD:
        x_lower = optimal_scaling_bounds[2 * x_category - 2]
        x_upper = optimal_scaling_bounds[2 * x_category - 1]
    else:
        raise NotImplementedError(
            f"Unknown optimal scaling method {options[METHOD]}. "
            f"Please use {REDUCED} or {STANDARD}."
        )
    return x_upper, x_lower


def get_constraints_for_optimization(
    xs: List[OptimalScalingParameter], sim: List[np.ndarray], options: Dict
) -> Dict:
    """Return constraints for inner optimization."""
    num_categories = len(xs)
    interval_range, interval_gap = compute_interval_constraints(
        xs, sim, options
    )
    if options[METHOD] == REDUCED:
        a = np.diag(-np.ones(num_categories), -1) + np.diag(
            np.ones(num_categories + 1)
        )
        a = a[:-1, :-1]
        b = np.empty((num_categories,))
        b[0] = interval_range
        b[1:] = interval_range + interval_gap
    elif options[METHOD] == STANDARD:
        a = np.diag(-np.ones(2 * num_categories), -1) + np.diag(
            np.ones(2 * num_categories + 1)
        )
        a = a[:-1, :-1]
        b = np.empty((2 * num_categories,))
        b[0] = 0
        b[1::2] = interval_range
        b[2::2] = interval_gap
    ineq_cons = {'type': 'ineq', 'fun': lambda x: a.dot(x) - b}

    return ineq_cons


def save_inner_parameters_to_inner_problem(
    category_upper_bounds: List[OptimalScalingParameter],
    group: int,
    problem: OptimalScalingProblem,
    x_inner_opt: Dict,
    sim: List[np.ndarray],
    options: Dict,
) -> None:
    """Save inner parameter values to the inner subproblem."""
    interval_range, interval_gap = compute_interval_constraints(
        category_upper_bounds, sim, options
    )

    surrogate_all, x_lower, x_upper = get_surrogate_all(
        category_upper_bounds,
        x_inner_opt[SCIPY_X],
        sim,
        interval_range,
        interval_gap,
        options,
    )
    problem.groups[group][SURROGATE_DATA] = surrogate_all.flatten()

    for inner_parameter in problem.get_cat_ub_parameters_for_group(group):
        inner_parameter.value = x_upper[inner_parameter.category - 1]
    for inner_parameter in problem.get_cat_lb_parameters_for_group(group):
        inner_parameter.value = x_lower[inner_parameter.category - 1]


def calculate_censored_obj(
    category_upper_bounds: List[OptimalScalingParameter],
    category_lower_bounds: List[OptimalScalingParameter],
    sim: List[np.ndarray],
    sigma: List[np.ndarray],
    quantitative_data: np.ndarray,
    quantitative_ixs: List[np.ndarray],
) -> Dict:
    """Calculate objective function for a group with censored data.

    Parameters
    ----------
    category_upper_bounds:
        The upper bounds for the categories.
    category_lower_bounds:
        The lower bounds for the categories.
    sim:
        The model simulation.
    sigma:
        The noise parameters from AMICI.
    observable_ids:
        The observable ids from the model.

    Returns
    -------
    Dictionary with the objective function value, dummy success
    and censoring category bounds.
    """
    cat_lb_values = np.array([x.value for x in category_lower_bounds])
    cat_ub_values = np.array([x.value for x in category_upper_bounds])

    obj = 0
    # Calculate the objective function for censored data.
    for cat_lb, cat_ub, cat_ub_par in zip(
        cat_lb_values,
        cat_ub_values,
        category_upper_bounds,
    ):
        for sim_i, sigma_i, mask_i in zip(sim, sigma, cat_ub_par.ixs):
            y_sim = sim_i[mask_i]
            if len(y_sim) == 0:
                continue
            sigma_for_observable = sigma_i[mask_i][0]
            for y_sim_i in y_sim:
                if cat_lb > y_sim_i:
                    y_surrogate = cat_lb
                elif y_sim_i > cat_ub:
                    y_surrogate = cat_ub
                else:
                    y_surrogate = y_sim_i
                obj += 0.5 * (
                    np.log(2 * np.pi * sigma_for_observable**2)
                    + (y_surrogate - y_sim_i) ** 2 / sigma_for_observable**2
                )

    # Gather the simulation and sigma values for the quantitative data.
    quantitative_sim = np.concatenate(
        [sim_i[mask_i] for sim_i, mask_i in zip(sim, quantitative_ixs)]
    )
    quantitative_sigma = np.concatenate(
        [sigma_i[mask_i] for sigma_i, mask_i in zip(sigma, quantitative_ixs)]
    )

    # Calculate the objective function for uncensored, quantitative data.
    obj += 0.5 * np.nansum(
        np.log(2 * np.pi * quantitative_sigma**2)
        + (quantitative_data - quantitative_sim) ** 2 / quantitative_sigma**2
    )

    return_dictionary = {
        SCIPY_SUCCESS: True,
        SCIPY_FUN: obj,
        SCIPY_X: np.ravel([cat_lb_values, cat_ub_values], order='F'),
    }
    return return_dictionary


def calculate_censored_grad(
    category_upper_bounds: List[OptimalScalingParameter],
    category_lower_bounds: List[OptimalScalingParameter],
    sim: List[np.ndarray],
    sy: np.ndarray,
    sigma: List[np.ndarray],
    ssigma: np.ndarray,
    par_edata_idx: List,
    quantitative_data: np.ndarray,
    quantitative_ixs: List[np.ndarray],
):
    """Calculate gradient for a group with censored data with respect to an outer parameter.

    Parameters
    ----------
    category_upper_bounds:
        The upper bounds for the categories.
    category_lower_bounds:
        The lower bounds for the categories.
    sim:
        The model simulation.
    sy_all:
        The sensitivities of the simulation.
    sigma:
        The noise parameters from AMICI.
    ssigma_all:
        The sensitivities of the noise parameters.

    Returns
    -------
    Gradient.
    """
    cat_lb_values = np.array([x.value for x in category_lower_bounds])
    cat_ub_values = np.array([x.value for x in category_upper_bounds])

    surrogate_all = []
    sim_all = []
    sigma_all = []
    sy_all = get_sy_all(category_upper_bounds, sy, par_edata_idx)
    ssigma_all = get_sy_all(category_upper_bounds, ssigma, par_edata_idx)

    # Gather the surrogate data, the simulation data
    # and the noise parameter arrays across categories.
    for cat_lb, cat_ub, cat_ub_par in zip(
        cat_lb_values,
        cat_ub_values,
        category_upper_bounds,
    ):
        for sim_i, sigma_i, mask_i in zip(sim, sigma, cat_ub_par.ixs):
            y_sim = sim_i[mask_i]
            if len(y_sim) == 0:
                continue
            sigma_for_observable = sigma_i[mask_i][0]
            for y_sim_i in y_sim:
                if cat_lb > y_sim_i:
                    y_surrogate = cat_lb
                elif y_sim_i > cat_ub:
                    y_surrogate = cat_ub
                else:
                    y_surrogate = y_sim_i
                sim_all.append(y_sim_i)
                sigma_all.append(sigma_for_observable)
                surrogate_all.append(y_surrogate)

    sim_all = np.array(sim_all)
    sigma_all = np.array(sigma_all)
    surrogate_all = np.array(surrogate_all)

    # Calculate the negative log likelihood gradient for censored data.
    gradient = np.nansum(
        (
            np.full(len(sim_all), 1)
            - (surrogate_all - sim_all) ** 2 / sigma_all**2
        )
        * ssigma_all
        / sigma_all
    ) - np.nansum((surrogate_all - sim_all) * sy_all / sigma_all**2)

    # Gather the simulation, sigma values and sensitivities for the quantitative data.
    quantitative_sim = np.concatenate(
        [sim_i[mask_i] for sim_i, mask_i in zip(sim, quantitative_ixs)]
    )
    quantitative_sigma = np.concatenate(
        [sigma_i[mask_i] for sigma_i, mask_i in zip(sigma, quantitative_ixs)]
    )
    quantitative_sy = np.concatenate(
        [
            sy_i[:, edata_idx, :][mask_i]
            if edata_idx is not None
            else np.full(sy_i[:, 0, :][mask_i].shape, 0)
            for sy_i, mask_i, edata_idx in zip(
                sy, quantitative_ixs, par_edata_idx
            )
        ]
    )
    quantitative_ssigma = np.concatenate(
        [
            ssigma_i[:, edata_idx, :][mask_i]
            if edata_idx is not None
            else np.full(ssigma_i[:, 0, :][mask_i].shape, 0)
            for ssigma_i, mask_i, edata_idx in zip(
                ssigma, quantitative_ixs, par_edata_idx
            )
        ]
    )

    # Calculate the negative log likelihood gradient for uncensored, quantitative data.
    gradient += np.nansum(
        (
            np.full(len(quantitative_sim), 1)
            - (quantitative_data - quantitative_sim) ** 2
            / quantitative_sigma**2
        )
        * quantitative_ssigma
        / quantitative_sigma
    ) - np.nansum(
        (quantitative_data - quantitative_sim)
        * quantitative_sy
        / quantitative_sigma**2
    )

    return gradient



"""Overall result."""

from .optimize import OptimizeResult
from .profile import ProfileResult
from .sample import SampleResult


class Result:
    """
    Universal result object for pypesto.

    The algorithms like optimize, profile,
    sample fill different parts of it.

    Attributes
    ----------
    problem: pypesto.Problem
        The problem underlying the results.
    optimize_result:
        The results of the optimizer runs.
    profile_result:
        The results of the profiler run.
    sample_result:
        The results of the sampler run.
    """

    def __init__(
        self,
        problem=None,
        optimize_result: OptimizeResult = None,
        profile_result: ProfileResult = None,
        sample_result: SampleResult = None,
    ):
        self.problem = problem
        self.optimize_result = optimize_result or OptimizeResult()
        self.profile_result = profile_result or ProfileResult()
        self.sample_result = sample_result or SampleResult()

    def summary(self, full: bool = False) -> str:
        """
        Get summary of the object.

        Parameters
        ----------
        full:
            If True, print full vectors including fixed parameters.
        """
        return self.optimize_result.summary(full=full)



import logging

from ...C import InnerParameterType
from ..parameter import InnerParameter

logger = logging.getLogger(__name__)


class SplineInnerParameter(InnerParameter):
    """A spline (inner) parameter of the spline hierarchical optimization problem.

    Attributes
    ----------
    observable_id:
        The id of the observable the spline is modeling.
    group:
        Group index. Corresponds to observable index + 1.
    index:
        Parameter index inside the group. Ranges from 1 to n_spline_parameters
        of its group.
    value:
        Current value of the inner parameter.
    estimate:
        Whether to estimate inner parameter in inner subproblem.
    """

    def __init__(
        self,
        *args,
        observable_id: str = None,
        group: int = None,
        index: int = None,
        estimate: bool = False,
        **kwargs,
    ):
        """Construct.

        Parameters
        ----------
        See class attributes.
        """
        super().__init__(*args, **kwargs)
        if self.inner_parameter_type != InnerParameterType.SPLINE:
            raise ValueError(
                "For the SplineParameter class, the parameter type has to be spline."
            )

        if observable_id is None:
            raise ValueError("No observable id provided.")
        if group is None:
            raise ValueError("No Parameter group provided.")
        if index is None:
            raise ValueError("No Parameter index provided.")

        self.observable_id = observable_id
        self.group = group
        self.index = index
        self.estimate = estimate
        self.value = self.dummy_value



"""Enhanced Scatter Search."""
from .cess import CESSOptimizer
from .ess import ESSOptimizer
from .function_evaluator import (
    FunctionEvaluator,
    FunctionEvaluatorMP,
    FunctionEvaluatorMT,
)
from .refset import RefSet
from .sacess import SacessOptimizer, get_default_ess_options
